/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.LinkedList;
import java.util.List;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.Path;

/**
 * This class implements an <a href="http://ant.apache.org">Ant</a> task
 * diassembling a class. <br>
 * The task supports the following attributes:
 * <ul>
 *   <li> <tt>classname</tt>: fully-qualified name of the class to disassemble<br>
 *        <i>string</i> - no default value, should be set
 *   <li> <tt>output</tt>: where to disassemble the class <br>
          <i>file</i> defaults to <tt>null</tt> (which means standard output)
 * </ul> <br>
 * The task should also be supplied a nested <tt>&lt;classpath&gt;</tt> element
 * to know where to look for class files.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public final class AntDisassembleTask extends Task {

    /** Fully-qualified name of the class to be disassembled. */
    private String classname;

    /** Where to disassemble the class (<tt>null</tt> means standard output). */
    private File output;

    /** Nested classpath elements. */
    private List<Path> classpaths;

    /**
     * Creates a new task with default parameters.
     */
    public AntDisassembleTask() {
        this.classname = null;
        this.output = null;
        this.classpaths = new LinkedList<Path>();
    } // end empty constructor

    /**
     * Sets the fully-qualified name of the class to disassemble.
     * @param cn fully-qualified name of the class to disassemble
     */
    public void setClassname(final String cn) {
        this.classname = cn;
    } // end method 'setClassname(String)'

    /**
     * Sets the output for the task.
     * @param f output for the task
     */
    public void setOutput(final File f) {
        this.output = f;
    } // end method 'setOutput(File)'

    /**
     * Adds a nested classpath element to the task.
     * @param p classpath element to add
     */
    public void addClasspath(final Path p) {
        if (p != null) {
            this.classpaths.add(p);
        } // end if
    } // end method 'addClasspath(Path)'

    /**
     * Actually executes the task.
     */
    public void execute() throws BuildException {
        if (this.classname == null) {
            throw new BuildException("'classname' attribute is not set.",
                                     getLocation());
        } // end if

        final List<String> classpath = new LinkedList<String>();
        for (Path p : this.classpaths) {
            for (String s : p.list()) {
                classpath.add(new File(s).getAbsolutePath());
            } // end for
        } // end for

        try {
            final OutputStream out =
                this.output == null ? System.out : new FileOutputStream(this.output);
            Disassembler.disassemble(this.classname, new ClassPath(classpath), out);
        } catch (final IOException ioe) {
            throw new BuildException(ioe.getMessage(), ioe, getLocation());
        } catch (final BaristaException be) {
            throw new BuildException(be.getMessage(), be, getLocation());
        } // end try/catch
    } // end method 'execute()'

} // end class 'AntDisassembleTask'
