/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.LinkedList;
import java.util.List;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class represents a Java class definition.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public final class ClassDefinition implements ToValue {

    /** Class flags. */
    private final List<AccessFlag> flags;

    /** Class name. */
    private final String name;

    /**
     * Class parent
     * (should be <tt>null</tt>only for <tt>java.lang.Object</tt>).
     */
    private final String parent;

    /** Class interfaces. */
    private final List<String> interfaces;

    /** Class fields. */
    private final List<Field> fields;

    /** Class methods. */
    private final List<Method> methods;

    /** Class attributes. */
    private final List<? extends Attribute> attributes;

    /**
     * Constructs a class definition.
     * @param af class flags - should not be <tt>null</tt>,
     *           should also be class flags
     * @param n class name - should not be <tt>null</tt>
     * @param p class parent (should be <tt>null</tt>only for <tt>java.lang.Object</tt>)
     * @param i class interfaces - should not be <tt>null</tt>
     * @param f class fields - should not be <tt>null</tt>
     * @param m class methods - should not be <tt>null</tt>
     * @param a class attributes - should not be <tt>null</tt>,
     *           should also be class attributes
     * @throws BaristaException if one of the flags is not a valid flag
     * @throws BaristaException if one of the attributes is not a valid attribute
     */
    public ClassDefinition(final List<AccessFlag> af,
                           final String n,
                           final String p,
                           final List<String> i,
                           final List<Field> f,
                           final List<Method> m,
                           final List<? extends Attribute> a)
        throws BaristaException {
        assert af != null : "null af";
        assert n != null : "null n";
        assert i != null : "null i";
        assert f != null : "null f";
        assert m != null : "null m";
        assert a != null : "null a";
        if (!AccessFlag.areClassFlags(af)) {
            throw new BaristaException("invalid flag");
        } // end if
        if (!Attribute.areClassAttributes(a)) {
            throw new BaristaException("invalid attribute");
        } // end if
        this.flags = new LinkedList<AccessFlag>(af);
        this.name = n;
        this.parent = p;
        this.interfaces = new LinkedList<String>(i);
        this.fields = new LinkedList<Field>(f);
        this.methods = new LinkedList<Method>(m);
        this.attributes = new LinkedList<Attribute>(a);
    } // end constructor(List<AccessFlag>, String, String, ...)

    /**
     * Returns the class flags.
     * @return the class flags
     */
    public List<AccessFlag> getFlags() {
        return new LinkedList<AccessFlag>(this.flags);
    } // end method 'getFlags()'

    /**
     * Returns the class name.
     * @return the class name
     */
    public String getName() {
        return this.name;
    } // end method 'getName()'

    /**
     * Returns the class parent.
     * @return the class parent
     */
    public String getParent() {
        return this.parent;
    } // end method 'getParent()'

    /**
     * Returns the class interfaces.
     * @return the class interfaces
     */
    public List<String> getInterfaces() {
        return new LinkedList<String>(this.interfaces);
    } // end method 'getInterfaces()'

    /**
     * Returns the class fields.
     * @return the class fields
     */
    public List<Field> getFields() {
        return new LinkedList<Field>(this.fields);
    } // end method 'getFields()'

    /**
     * Returns the class methods.
     * @return the class methods
     */
    public List<Method> getMethods() {
        return new LinkedList<Method>(this.methods);
    } // end method 'getMethods()'

    /**
     * Returns the class attributes.
     * @return the class attributes
     */
    public List<? extends Attribute> getAttributes() {
        return new LinkedList<Attribute>(this.attributes);
    } // end method 'getAttributes()'

    /**
     * {@inheritDoc}
     */
    public Value toValue() {
        final Block b = Block.createBlock(7, 0);
        b.set(0, API.encodeList(this.flags));
        b.set(1, API.createClassName(this.name));
        if (this.parent == null) {
            b.set(2, Value.ZERO);
        } else {
            final Block p = Block.createBlock(0, API.createClassName(this.parent));
            b.set(2, Value.createFromBlock(p));
        } // end if/else
        final List<Value> l = new LinkedList<Value>();
        for (String s : this.interfaces) {
            l.add(API.createClassName(s));
        } // end for
        b.set(3, API.encodeValueList(l));
        b.set(4, API.encodeList(this.fields));
        b.set(5, API.encodeList(this.methods));
        b.set(6, API.encodeList(this.attributes));
        return Value.createFromBlock(b);
    } // end method 'toValue()'

    /**
     * Constructs a class definition from a value.
     * @param v value to constructs class definition from - should not be <tt>null</tt>
     * @return the class definition corresponding to the passed value
     */
    public static ClassDefinition fromValue(final Value v) {
        assert v != null : "null v";
        final Block b = v.asBlock();
        final List<AccessFlag> flags = new LinkedList<AccessFlag>();
        for (Value a : API.decodeList(b.get(0))) {
            flags.add(AccessFlag.fromValue(a));
        } // end for
        final List<String> itf = new LinkedList<String>();
        for (Value i : API.decodeList(b.get(3))) {
            itf.add(API.stringOfClassName(i));
        } // end for
        final List<Field> fld = new LinkedList<Field>();
        for (Value f : API.decodeList(b.get(4))) {
            fld.add(Field.fromValue(f));
        } // end for
        final List<Method> mth = new LinkedList<Method>();
        for (Value m : API.decodeList(b.get(5))) {
            mth.add(Method.fromValue(m));
        } // end for
        final List<Attribute> att = new LinkedList<Attribute>();
        for (Value a : API.decodeList(b.get(6))) {
            att.add(Attribute.fromValue(a));
        } // end for
        return new ClassDefinition(flags,
                                   API.stringOfClassName(b.get(1)),
                                   b.get(2) == Value.ZERO
                                   ? null
                                   : API.stringOfClassName(b.get(2).asBlock().get(0)),
                                   itf,
                                   fld,
                                   mth,
                                   att);
    } // end method 'fromValue(Value)'

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder();
        final String fl = AccessFlag.listToString(this.flags);
        if (fl.length() > 0) {
            sb.append(fl);
            sb.append(" ");
        } // end if
        sb.append("class ");
        sb.append(this.name);
        if (this.parent != null) {
            sb.append(" extends ");
            sb.append(this.parent);
        } // end for
        if (this.interfaces.size() > 0) {
            sb.append(" implements ");
            boolean first = true;
            for (String s : this.interfaces) {
                if (first) {
                    first = false;
                } else {
                    sb.append(", ");
                } // end if/else
                sb.append(s);
            } // end for
        } // end if
        sb.append(" {\n");
        for (Field f : this.fields) {
            sb.append("    ");
            sb.append(f.toString());
            sb.append(";\n");
        } // end for
        for (Method m : this.methods) {
            sb.append("    ");
            sb.append(m.toString());
            sb.append(";\n");
        } // end for
        sb.append("}");
        return sb.toString();
    } // end method 'toString()'

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return this.flags.hashCode()
            + this.name.hashCode()
            + (this.parent != null ? this.parent.hashCode() : 0)
            + this.interfaces.hashCode()
            + this.fields.hashCode()
            + this.methods.hashCode()
            + this.attributes.hashCode();
    } // end method 'hashCode()'

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof ClassDefinition) {
            final ClassDefinition that = (ClassDefinition) obj;
            return this.flags.equals(that.flags)
                && this.name.equals(that.name)
                && API.nullOrEquals(this.parent, that.parent)
                && this.interfaces.equals(that.interfaces)
                && this.fields.equals(that.fields)
                && this.methods.equals(that.methods)
                && this.attributes.equals(that.attributes);
        } else {
            return false;
        } // end if/else
    } // end method 'equals(Object)'
    
} // end class 'ClassDefinition'
