/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.stdlib.Hash;
import fr.x9c.cadmium.support.Helper;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class implements the various Java instructions.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.2
 * @since 1.0
 */
public abstract class Instruction implements ToValue {

    /**
     * Ensures that no instruction could be defined outside this file.
     */
    private Instruction() {
    } // end empty constructor

    /* This class represents a reference to a class or interface name. */
    public static final class TypeName implements ToValue {

        /** Class name. */
        private final String className;

        /**
         * Constructs an instance from a class or interface name.
         * @param n class or interface name - should not be <tt>null</tt>
         */
        public TypeName(final String n) {
            assert n != null : "null n";
            this.className = n;
        } // end constructor (String)

        /**
         * Returns the class or interface name.
         * @return the class or interface name
         */
        public String getClassName() {
            return this.className;
        } // end method 'getClassName()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return this.className;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.className.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof TypeName) {
                final TypeName that = (TypeName) obj;
                return this.className.equals(that.className);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            return API.createClassName(this.className);
        } // end method 'toValue()'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        static TypeName fromValue(final Value v) {
            return new TypeName(API.stringOfClassName(v));
        } // end method 'fromValue(Value)'

    } // end inner-class 'TypeName'

    /* This class represents a reference to a dynamic method. */
    public static final class DynamicMethodRef implements ToValue {

        /** Method name. */
        private final String methodName;

        /** Method return type. */
        private final Descriptor returnType;

        /** Method parameter types. */
        private final List<? extends Descriptor> parameterTypes;

        /**
         * Constructs a method reference from name, and return and parameter
         * types.
         * @param n method name - should not be <tt>null</tt>
         * @param r method return type - should not be <tt>null</tt>
         * @param p method parameter types - should not be <tt>null</tt>
         */
        public DynamicMethodRef(final String n,
                                final Descriptor r,
                                final List<? extends Descriptor> p) {
            assert n != null : "null n";
            assert r != null : "null r";
            assert p != null : "null p";
            this.methodName = n;
            this.returnType = r;
            this.parameterTypes = new LinkedList<Descriptor>(p);
        } // end constructor (String, Descriptor, List<? extends Descriptor>)

        /**
         * Returns the method name.
         * @return the method name
         */
        public String getMethodName() {
            return this.methodName;
        } // end method 'getMethodName()'

        /**
         * Returns the method return type.
         * @return the method return type
         */
        public Descriptor getReturnType() {
            return this.returnType;
        } // end method 'getReturnType()'

        /**
         * Returns the method parameter types.
         * @return the method parameter types
         */
        public List<? extends Descriptor> getParameterTypes() {
            return new LinkedList<Descriptor>(this.parameterTypes);
        } // end method 'getParameterTypes()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            final StringBuilder sb = new StringBuilder();
            sb.append(this.returnType.toString());
            sb.append(" ");
            sb.append(this.methodName);
            sb.append("(");
            sb.append(Descriptor.listToString(this.parameterTypes));
            sb.append(")");
            return sb.toString();
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.methodName.hashCode()
                 + this.returnType.hashCode()
                 + this.parameterTypes.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof DynamicMethodRef) {
                final DynamicMethodRef that = (DynamicMethodRef) obj;
                return this.methodName.equals(that.methodName)
                    && this.returnType.equals(that.returnType)
                    && this.parameterTypes.equals(that.parameterTypes);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block desc = Block.createBlock(0,
                                                 API.encodeList(this.parameterTypes),
                                                 this.returnType.toValue());
            final Block b = Block.createBlock(0,
                                              API.createMethodName(this.methodName),
                                              Value.createFromBlock(desc));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        private static DynamicMethodRef fromValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            final Descriptor r = Descriptor.fromValue(b.get(1).asBlock().get(1));
            final List<Descriptor> p = new LinkedList<Descriptor>();
            for (Value d : API.decodeList(b.get(1).asBlock().get(0))) {
                p.add(Descriptor.fromValue(d));
            } // end for
            return new DynamicMethodRef(API.stringOfMethodName(b.get(0)), r, p);
        } // end method 'fromValue(Value)'

    } // end inner-class 'DynamicMethodRef'

    /* This class represents a reference to a method. */
    public static final class MethodRef implements ToValue {

        /** Class name. */
        private final String className;

        /** Method name. */
        private final String methodName;

        /** Method return type. */
        private final Descriptor returnType;

        /** Method parameter types. */
        private final List<? extends Descriptor> parameterTypes;

        /**
         * Constructs a method reference from class and method names,
         * and return and parameter types.
         * @param c class name - should not be <tt>null</tt>
         * @param n method name - should not be <tt>null</tt>
         * @param r method return type - should not be <tt>null</tt>
         * @param p method parameter types - should not be <tt>null</tt>
         */
        public MethodRef(final String c,
                         final String n,
                         final Descriptor r,
                         final List<? extends Descriptor> p) {
            assert c != null : "null c";
            assert n != null : "null n";
            assert r != null : "null r";
            assert p != null : "null p";
            this.className = c;
            this.methodName = n;
            this.returnType = r;
            this.parameterTypes = new LinkedList<Descriptor>(p);
        } // end constructor (String, String, Descriptor, List<? extends Descriptor>)

        /**
         * Returns the class name.
         * @return the class name
         */
        public String getClassName() {
            return this.className;
        } // end method 'getClassName()'

        /**
         * Returns the method name.
         * @return the method name
         */
        public String getMethodName() {
            return this.methodName;
        } // end method 'getMethodName()'

        /**
         * Returns the method return type.
         * @return the method return type
         */
        public Descriptor getReturnType() {
            return this.returnType;
        } // end method 'getReturnType()'

        /**
         * Returns the method parameter types.
         * @return the method parameter types
         */
        public List<? extends Descriptor> getParameterTypes() {
            return new LinkedList<Descriptor>(this.parameterTypes);
        } // end method 'getParameterTypes()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            final StringBuilder sb = new StringBuilder();
            sb.append(this.returnType.toString());
            sb.append(" ");
            sb.append(this.className);
            sb.append(".");
            sb.append(this.methodName);
            sb.append("(");
            sb.append(Descriptor.listToString(this.parameterTypes));
            sb.append(")");
            return sb.toString();
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.className.hashCode()
                 + this.methodName.hashCode()
                 + this.returnType.hashCode()
                 + this.parameterTypes.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof MethodRef) {
                final MethodRef that = (MethodRef) obj;
                return this.className.equals(that.className)
                    && this.methodName.equals(that.methodName)
                    && this.returnType.equals(that.returnType)
                    && this.parameterTypes.equals(that.parameterTypes);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block desc = Block.createBlock(0,
                                                 API.encodeList(this.parameterTypes),
                                                 this.returnType.toValue());
            final Block b = Block.createBlock(0,
                                              API.createClassName(this.className),
                                              API.createMethodName(this.methodName),
                                              Value.createFromBlock(desc));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * Same as {@link #toValue()}, but building a tagged value.
         * @return the instance, as a tagged value
         */
        public Value toTaggedValue() {
            final Block desc = Block.createBlock(0,
                                                 API.encodeList(this.parameterTypes),
                                                 this.returnType.toValue());
            final Block b = Block.createBlock(0,
                                              wrap("Class_or_interface", API.createClassName(this.className)),
                                              API.createMethodName(this.methodName),
                                              Value.createFromBlock(desc));
            return Value.createFromBlock(b);
        } // end method 'toTaggedValue()'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        private static MethodRef fromValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            final Descriptor r = Descriptor.fromValue(b.get(2).asBlock().get(1));
            final List<Descriptor> p = new LinkedList<Descriptor>();
            for (Value d : API.decodeList(b.get(2).asBlock().get(0))) {
                p.add(Descriptor.fromValue(d));
            } // end for
            return new MethodRef(API.stringOfClassName(b.get(0)),
                                 API.stringOfMethodName(b.get(1)),
                                 r,
                                 p);
        } // end method 'fromValue(Value)'

        /**
         * Constructs an instance from a (tagged) value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        private static MethodRef fromTaggedValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            final Descriptor r = Descriptor.fromValue(b.get(2).asBlock().get(1));
            final List<Descriptor> p = new LinkedList<Descriptor>();
            for (Value d : API.decodeList(b.get(2).asBlock().get(0))) {
                p.add(Descriptor.fromValue(d));
            } // end for
            return new MethodRef(API.stringOfClassName(b.get(0).asBlock().get(1)),
                                 API.stringOfMethodName(b.get(1)),
                                 r,
                                 p);
        } // end method 'fromTaggedValue(Value)'

    } // end inner-class 'MethodRef'

    /* This class represents a reference to an array method. */
    public static final class ArrayMethodRef implements ToValue {

        /** Array type. */
        private final Descriptor.Array arrayType;

        /** Method name. */
        private final String methodName;

        /** Method return type. */
        private final Descriptor returnType;

        /** Method parameter types. */
        private final List<? extends Descriptor> parameterTypes;

        /**
         * Constructs a method reference from class and method names,
         * and return and parameter types.
         * @param a array type - should not be <tt>null</tt>
         * @param n method name - should not be <tt>null</tt>
         * @param r method return type - should not be <tt>null</tt>
         * @param p method parameter types - should not be <tt>null</tt>
         */
        public ArrayMethodRef(final Descriptor.Array a,
                              final String n,
                              final Descriptor r,
                              final List<? extends Descriptor> p) {
            assert a != null : "null a";
            assert n != null : "null n";
            assert r != null : "null r";
            assert p != null : "null p";
            this.arrayType = a;
            this.methodName = n;
            this.returnType = r;
            this.parameterTypes = new LinkedList<Descriptor>(p);
        } // end constructor (Descriptor.Array, String, Descriptor, List<? extends Descriptor>)

        /**
         * Returns the array type.
         * @return the array type
         */
        public Descriptor.Array getArrayType() {
            return this.arrayType;
        } // end method getArrayType()'

        /**
         * Returns the method name.
         * @return the method name
         */
        public String getMethodName() {
            return this.methodName;
        } // end method 'getMethodName()'

        /**
         * Returns the method return type.
         * @return the method return type
         */
        public Descriptor getReturnType() {
            return this.returnType;
        } // end method 'getReturnType()'

        /**
         * Returns the method parameter types.
         * @return the method parameter types
         */
        public List<? extends Descriptor> getParameterTypes() {
            return new LinkedList<Descriptor>(this.parameterTypes);
        } // end method 'getParameterTypes()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            final StringBuilder sb = new StringBuilder();
            sb.append(this.returnType.toString());
            sb.append(" ");
            sb.append(this.arrayType);
            sb.append(".");
            sb.append(this.methodName);
            sb.append("(");
            sb.append(Descriptor.listToString(this.parameterTypes));
            sb.append(")");
            return sb.toString();
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.arrayType.hashCode()
                 + this.methodName.hashCode()
                 + this.returnType.hashCode()
                 + this.parameterTypes.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ArrayMethodRef) {
                final ArrayMethodRef that = (ArrayMethodRef) obj;
                return this.arrayType.equals(that.arrayType)
                    && this.methodName.equals(that.methodName)
                    && this.returnType.equals(that.returnType)
                    && this.parameterTypes.equals(that.parameterTypes);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block desc = Block.createBlock(0,
                                                 API.encodeList(this.parameterTypes),
                                                 this.returnType.toValue());
            final Block b = Block.createBlock(0,
                                              this.arrayType.toValue(),
                                              API.createMethodName(this.methodName),
                                              Value.createFromBlock(desc));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * Same as {@link #toValue()}, but building a tagged value.
         * @return the instance, as a tagged value
         */
        public Value toTaggedValue() {
            final Block desc = Block.createBlock(0,
                                                 API.encodeList(this.parameterTypes),
                                                 this.returnType.toValue());
            final Block b = Block.createBlock(0,
                                              wrap("Array_type", this.arrayType.toValue()),
                                              API.createMethodName(this.methodName),
                                              Value.createFromBlock(desc));
            return Value.createFromBlock(b);
        } // end method 'toTaggedValue()'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        private static ArrayMethodRef fromValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            final Descriptor r = Descriptor.fromValue(b.get(2).asBlock().get(1));
            final List<Descriptor> p = new LinkedList<Descriptor>();
            for (Value d : API.decodeList(b.get(2).asBlock().get(0))) {
                p.add(Descriptor.fromValue(d));
            } // end for
            return new ArrayMethodRef((Descriptor.Array) Descriptor.fromValue(b.get(0)),
                                 API.stringOfMethodName(b.get(1)),
                                 r,
                                 p);
        } // end method 'fromValue(Value)'

        /**
         * Constructs an instance from a (tagged) value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        private static ArrayMethodRef fromTaggedValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            final Descriptor r = Descriptor.fromValue(b.get(2).asBlock().get(1));
            final List<Descriptor> p = new LinkedList<Descriptor>();
            for (Value d : API.decodeList(b.get(2).asBlock().get(0))) {
                p.add(Descriptor.fromValue(d));
            } // end for
            return new ArrayMethodRef((Descriptor.Array) Descriptor.fromValue(b.get(0).asBlock().get(1)),
                                 API.stringOfMethodName(b.get(1)),
                                 r,
                                 p);
        } // end method 'fromTaggedValue(Value)'

    } // end inner-class 'ArrayMethodRef'

    /* This class represents a reference to a field. */
    public static final class FieldRef implements ToValue {

        /** Class name. */
        private final String className;

        /** Field name. */
        private final String fieldName;

        /** Field type. */
        private final Descriptor fieldType;

        /**
         * Constructs and instance from class name, and field name and descriptor.
         * @param cn class name - should not be <tt>null</tt>
         * @param fn field name - should not be <tt>null</tt>
         * @param ft field type - should not be <tt>null</tt>
         * @throws BaristaException if the descriptor is not a valid descriptor
         */
        public FieldRef(final String cn, final String fn, final Descriptor ft)
            throws BaristaException {
            assert cn != null : "null cn";
            assert fn != null : "null fn";
            assert ft != null : "null ft";
            if (!ft.isFieldDescriptor()) {
                throw new BaristaException("invalid descriptor");
            } // end if
            this.className = cn;
            this.fieldName = fn;
            this.fieldType = ft;
        } // end constructor (String, String, Descriptor)

        /**
         * Returns the class name.
         * @return the class name
         */
        public String getClassName() {
            return this.className;
        } // end method 'getClassName()'

        /**
         * Returns the field name.
         * @return the field name
         */
        public String getFieldName() {
            return this.fieldName;
        } // end method 'getFieldName()'

        /**
         * Returns the field type.
         * @return the field type
         */
        public Descriptor getFieldType() {
            return this.fieldType;
        } // end method 'getFieldType()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            final StringBuilder sb = new StringBuilder();
            sb.append(this.fieldType.toString());
            sb.append(" ");
            sb.append(this.className);
            sb.append(".");
            sb.append(this.fieldName);
            return sb.toString();
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.className.hashCode()
                 + this.fieldName.hashCode()
                 + this.fieldType.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof FieldRef) {
                final FieldRef that = (FieldRef) obj;
                return this.className.equals(that.className)
                    && this.fieldName.equals(that.fieldName)
                    && this.fieldType.equals(that.fieldType);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              API.createClassName(this.className),
                                              API.createFieldName(this.fieldName),
                                              this.fieldType.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        private static FieldRef fromValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            return new FieldRef(API.stringOfClassName(b.get(0)),
                                API.stringOfFieldName(b.get(1)),
                                Descriptor.fromValue(b.get(2)));
        } // end method 'fromValue(Value)'

    } // end inner-class 'FieldRef'

    /**
     * This class implements a match/offset pair.
     */
    public static final class MatchOffsetPair implements ToValue {

        /** The match value. */
        private final int match;

        /** The offset value. */
        private final int offset;

        /**
         * Constructs a match/offset pair.
         * @param m match value
         * @param o offset value
         */
        public MatchOffsetPair(final int m, final int o) {
            this.match = m;
            this.offset = o;
        } // end constructor (int, int)

        /**
         * Returns the match value.
         * @return the match value
         */
        public int getMatch() {
            return this.match;
        } // end method 'getMatch()'

        /**
         * Returns the offset value.
         * @return the offset value
         */
        public int getOffset() {
            return this.offset;
        } // end method 'getOffset()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return this.match + " => " + this.offset;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.match + this.offset;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof MatchOffsetPair) {
                final MatchOffsetPair that = (MatchOffsetPair) obj;
                return this.match == that.match
                    && this.offset == that.offset;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Value m = Helper.createInt32(this.match);
            final Value o = Helper.createInt32(this.offset);
            return Value.createFromBlock(Block.createBlock(0, m, o));
        } // end method 'toValue()'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        private static MatchOffsetPair fromValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            return new MatchOffsetPair(b.get(0).asBlock().asInt32(),
                                       b.get(1).asBlock().asInt32());
        } // end method 'fromValue(Value)'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        private static List<MatchOffsetPair> listFromValue(final Value v) {
            assert v != null : "null v";
            final List<Value> l = API.decodeList(v);
            final List<MatchOffsetPair> res = new LinkedList<MatchOffsetPair>();
            for (Value x : l) {
                res.add(fromValue(x));
            } // end for
            return res;
        } // end method 'listFromValue(Value)'

    } // end inner-class 'MatchOffsetPair'

    /**
     * Decodes a list of integers from a value.
     * @param v value to decode - should not be <tt>null</tt>
     * @return the corresponding list
     */
    private static List<Integer> readIntegerList(final Value v) {
        assert v != null : "null v";
        final List<Integer> res = new LinkedList<Integer>();
        Value list = v;
        while (list != Value.EMPTY_LIST) {
            final Block b = list.asBlock();
            res.add(b.get(0).asBlock().asInt32());
            list = b.get(1);
        } // end while
        return res;
    } // end method 'readIntegerList(Value)'

    /**
     * Encodes a list of integers into a value.
     * @param l list to encode - should not be <tt>null</tt>
     * @return the passed list as a value
     */
    private static Value encodeIntegerList(final List<Integer> l) {
        assert l != null : "null l";
        Value res = Value.EMPTY_LIST;
        final ListIterator<Integer> it = l.listIterator(l.size());
        while (it.hasPrevious()) {
            final Block cons = Block.createBlock(Block.TAG_CONS,
                                                 Helper.createInt32(it.previous()),
                                                 res);
            res = Value.createFromBlock(cons);
        } // end while
        return res;
    } // end method 'encodeIntegerList(final List<Integer>)'

    /**
     * Wraps a value inside a variant.
     * @param id variant identifier - should not be <tt>null</tt>
     * @param v value to wrap - should not be <tt>null</tt>
     * @return the wrapped value
     */
    static Value wrap(final String id, final Value v) {
        assert id != null : "null id";
        assert v != null : "null v";
        final Block b = Block.createBlock(0, Hash.hashVariant(id), v);
        return Value.createFromBlock(b);
    } // end method 'wrap(String, Value)'

    /**
     * This class represents instructions with no parameter.
     */
    private static final class Simple extends Instruction {

        /** Instruction name. */
        private final String name;

        /** Instruction code. */
        private final int code;

        /**
         * Constructs an instruction from name and code.
         * @param n instruction name - should not be <tt>null</tt>
         * @param c instruction code
         */
        private Simple(final String n, final int c) {
            assert n != null : "null n";
            this.name = n;
            this.code = c;
        } // end constructor(String, int)

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            return Value.createFromLong(this.code);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return this.name;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.code;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof Simple) {
                final Simple that = (Simple) obj;
                return this.code == that.code;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'Simple'

    /** Defines the AALOAD instruction (no parameter): load reference from array. */
    public static final Instruction AALOAD = new Simple("AALOAD", 0);

    /** Defines the AASTORE instruction (no parameter): store into reference array. */
    public static final Instruction AASTORE = new Simple("AASTORE", 1);

    /** Defines the ACONST_NULL instruction (no parameter): push null. */
    public static final Instruction ACONST_NULL = new Simple("ACONST_NULL", 2);

    /** Defines the ALOAD_0 instruction (no parameter): load reference from local variable. */
    public static final Instruction ALOAD_0 = new Simple("ALOAD_0", 3);

    /** Defines the ALOAD_1 instruction (no parameter): load reference from local variable. */
    public static final Instruction ALOAD_1 = new Simple("ALOAD_1", 4);

    /** Defines the ALOAD_2 instruction (no parameter): load reference from local variable. */
    public static final Instruction ALOAD_2 = new Simple("ALOAD_2", 5);

    /** Defines the ALOAD_3 instruction (no parameter): load reference from local variable. */
    public static final Instruction ALOAD_3 = new Simple("ALOAD_3", 6);

    /** Defines the ARETURN instruction (no parameter): return reference from method. */
    public static final Instruction ARETURN = new Simple("ARETURN", 7);

    /** Defines the ARRAYLENGTH instruction (no parameter): get length of array. */
    public static final Instruction ARRAYLENGTH = new Simple("ARRAYLENGTH", 8);

    /** Defines the ASTORE_0 instruction (no parameter): store reference into local variable. */
    public static final Instruction ASTORE_0 = new Simple("ASTORE_0", 9);

    /** Defines the ASTORE_1 instruction (no parameter): store reference into local variable. */
    public static final Instruction ASTORE_1 = new Simple("ASTORE_1", 10);

    /** Defines the ASTORE_2 instruction (no parameter): store reference into local variable. */
    public static final Instruction ASTORE_2 = new Simple("ASTORE_2", 11);

    /** Defines the ASTORE_3 instruction (no parameter): store reference into local variable. */
    public static final Instruction ASTORE_3 = new Simple("ASTORE_3", 12);

    /** Defines the ATHROW instruction (no parameter): throw exception or error. */
    public static final Instruction ATHROW = new Simple("ATHROW", 13);

    /** Defines the BALOAD instruction (no parameter): load byte or boolean from array. */
    public static final Instruction BALOAD = new Simple("BALOAD", 14);

    /** Defines the BASTORE instruction (no parameter): store into byte or boolean array. */
    public static final Instruction BASTORE = new Simple("BASTORE", 15);

    /** Defines the CALOAD instruction (no parameter): load char from array. */
    public static final Instruction CALOAD = new Simple("CALOAD", 16);

    /** Defines the CASTORE instruction (no parameter): store into char array. */
    public static final Instruction CASTORE = new Simple("CASTORE", 17);

    /** Defines the D2F instruction (no parameter): convert double to float. */
    public static final Instruction D2F = new Simple("D2F", 18);

    /** Defines the D2I instruction (no parameter): convert double to int. */
    public static final Instruction D2I = new Simple("D2I", 19);

    /** Defines the D2L instruction (no parameter): convert double to long. */
    public static final Instruction D2L = new Simple("D2L", 20);

    /** Defines the DADD instruction (no parameter): add double. */
    public static final Instruction DADD = new Simple("DADD", 21);

    /** Defines the DALOAD instruction (no parameter): load double from array. */
    public static final Instruction DALOAD = new Simple("DALOAD", 22);

    /** Defines the DASTORE instruction (no parameter): store into double array. */
    public static final Instruction DASTORE = new Simple("DASTORE", 23);

    /** Defines the DCMPG instruction (no parameter): compare double. */
    public static final Instruction DCMPG = new Simple("DCMPG", 24);

    /** Defines the DCMPL instruction (no parameter): compare double. */
    public static final Instruction DCMPL = new Simple("DCMPL", 25);

    /** Defines the DCONST_0 instruction (no parameter): push double. */
    public static final Instruction DCONST_0 = new Simple("DCONST_0", 26);

    /** Defines the DCONST_1 instruction (no parameter): push double. */
    public static final Instruction DCONST_1 = new Simple("DCONST_1", 27);

    /** Defines the DDIV instruction (no parameter): divide double. */
    public static final Instruction DDIV = new Simple("DDIV", 28);

    /** Defines the DLOAD_0 instruction (no parameter): load double from local variable. */
    public static final Instruction DLOAD_0 = new Simple("DLOAD_0", 29);

    /** Defines the DLOAD_1 instruction (no parameter): load double from local variable. */
    public static final Instruction DLOAD_1 = new Simple("DLOAD_1", 30);

    /** Defines the DLOAD_2 instruction (no parameter): load double from local variable. */
    public static final Instruction DLOAD_2 = new Simple("DLOAD_2", 31);

    /** Defines the DLOAD_3 instruction (no parameter): load double from local variable. */
    public static final Instruction DLOAD_3 = new Simple("DLOAD_3", 32);

    /** Defines the DMUL instruction (no parameter): multiply double. */
    public static final Instruction DMUL = new Simple("DMUL", 33);

    /** Defines the DNEG instruction (no parameter): negate double. */
    public static final Instruction DNEG = new Simple("DNEG", 34);

    /** Defines the DREM instruction (no parameter): remainder double. */
    public static final Instruction DREM = new Simple("DREM", 35);

    /** Defines the DRETURN instruction (no parameter): return double from method. */
    public static final Instruction DRETURN = new Simple("DRETURN", 36);

    /** Defines the DSTORE_0 instruction (no parameter): store double into local variable. */
    public static final Instruction DSTORE_0 = new Simple("DSTORE_0", 37);

    /** Defines the DSTORE_1 instruction (no parameter): store double into local variable. */
    public static final Instruction DSTORE_1 = new Simple("DSTORE_1", 38);

    /** Defines the DSTORE_2 instruction (no parameter): store double into local variable. */
    public static final Instruction DSTORE_2 = new Simple("DSTORE_2", 39);

    /** Defines the DSTORE_3 instruction (no parameter): store double into local variable. */
    public static final Instruction DSTORE_3 = new Simple("DSTORE_3", 40);

    /** Defines the DSUB instruction (no parameter): subtract double. */
    public static final Instruction DSUB = new Simple("DSUB", 41);

    /** Defines the DUP instruction (no parameter): duplicate the top operand stack value. */
    public static final Instruction DUP = new Simple("DUP", 42);

    /** Defines the DUP2 instruction (no parameter): duplicate the top one or two operand stack values. */
    public static final Instruction DUP2 = new Simple("DUP2", 43);

    /** Defines the DUP2_X1 instruction (no parameter): duplicate the top one or two operand stack values and insert two or three values down. */
    public static final Instruction DUP2_X1 = new Simple("DUP2_X1", 44);

    /** Defines the DUP2_X2 instruction (no parameter): duplicate the top one or two operand stack values and insert two, three, or four values down. */
    public static final Instruction DUP2_X2 = new Simple("DUP2_X2", 45);

    /** Defines the DUP_X1 instruction (no parameter): duplicate the top operand stack value and insert two values down. */
    public static final Instruction DUP_X1 = new Simple("DUP_X1", 46);

    /** Defines the DUP_X2 instruction (no parameter): duplicate the top operand stack value and insert two or threee values down. */
    public static final Instruction DUP_X2 = new Simple("DUP_X2", 47);

    /** Defines the F2D instruction (no parameter): convert float to double. */
    public static final Instruction F2D = new Simple("F2D", 48);

    /** Defines the F2I instruction (no parameter): convert float to int. */
    public static final Instruction F2I = new Simple("F2I", 49);

    /** Defines the F2L instruction (no parameter): convert float to long. */
    public static final Instruction F2L = new Simple("F2L", 50);

    /** Defines the FADD instruction (no parameter): add float. */
    public static final Instruction FADD = new Simple("FADD", 51);

    /** Defines the FALOAD instruction (no parameter): load float from array. */
    public static final Instruction FALOAD = new Simple("FALOAD", 52);

    /** Defines the FASTORE instruction (no parameter): store into float array. */
    public static final Instruction FASTORE = new Simple("FASTORE", 53);

    /** Defines the FCMPG instruction (no parameter): compare float. */
    public static final Instruction FCMPG = new Simple("FCMPG", 54);

    /** Defines the FCMPL instruction (no parameter): compare float. */
    public static final Instruction FCMPL = new Simple("FCMPL", 55);

    /** Defines the FCONST_0 instruction (no parameter): push float. */
    public static final Instruction FCONST_0 = new Simple("FCONST_0", 56);

    /** Defines the FCONST_1 instruction (no parameter): push float. */
    public static final Instruction FCONST_1 = new Simple("FCONST_1", 57);

    /** Defines the FCONST_2 instruction (no parameter): push float. */
    public static final Instruction FCONST_2 = new Simple("FCONST_2", 58);

    /** Defines the FDIV instruction (no parameter): divide float. */
    public static final Instruction FDIV = new Simple("FDIV", 59);

    /** Defines the FLOAD_0 instruction (no parameter): load float from local variable. */
    public static final Instruction FLOAD_0 = new Simple("FLOAD_0", 60);

    /** Defines the FLOAD_1 instruction (no parameter): load float from local variable. */
    public static final Instruction FLOAD_1 = new Simple("FLOAD_1", 61);

    /** Defines the FLOAD_2 instruction (no parameter): load float from local variable. */
    public static final Instruction FLOAD_2 = new Simple("FLOAD_2", 62);

    /** Defines the FLOAD_3 instruction (no parameter): load float from local variable. */
    public static final Instruction FLOAD_3 = new Simple("FLOAD_3", 63);

    /** Defines the FMUL instruction (no parameter): multiply float. */
    public static final Instruction FMUL = new Simple("FMUL", 64);

    /** Defines the FNEG instruction (no parameter): negate float. */
    public static final Instruction FNEG = new Simple("FNEG", 65);

    /** Defines the FREM instruction (no parameter): remainder float. */
    public static final Instruction FREM = new Simple("FREM", 66);

    /** Defines the FRETURN instruction (no parameter): return float from method. */
    public static final Instruction FRETURN = new Simple("FRETURN", 67);

    /** Defines the FSTORE_0 instruction (no parameter): store float into local variable. */
    public static final Instruction FSTORE_0 = new Simple("FSTORE_0", 68);

    /** Defines the FSTORE_1 instruction (no parameter): store float into local variable. */
    public static final Instruction FSTORE_1 = new Simple("FSTORE_1", 69);

    /** Defines the FSTORE_2 instruction (no parameter): store float into local variable. */
    public static final Instruction FSTORE_2 = new Simple("FSTORE_2", 70);

    /** Defines the FSTORE_3 instruction (no parameter): store float into local variable. */
    public static final Instruction FSTORE_3 = new Simple("FSTORE_3", 71);

    /** Defines the FSUB instruction (no parameter): subtract float. */
    public static final Instruction FSUB = new Simple("FSUB", 72);

    /** Defines the I2B instruction (no parameter): convert int to byte. */
    public static final Instruction I2B = new Simple("I2B", 73);

    /** Defines the I2C instruction (no parameter): convert int to char. */
    public static final Instruction I2C = new Simple("I2C", 74);

    /** Defines the I2D instruction (no parameter): convert int to double. */
    public static final Instruction I2D = new Simple("I2D", 75);

    /** Defines the I2F instruction (no parameter): convert int to float. */
    public static final Instruction I2F = new Simple("I2F", 76);

    /** Defines the I2L instruction (no parameter): convert int to long. */
    public static final Instruction I2L = new Simple("I2L", 77);

    /** Defines the I2S instruction (no parameter): convert int to short. */
    public static final Instruction I2S = new Simple("I2S", 78);

    /** Defines the IADD instruction (no parameter): add int. */
    public static final Instruction IADD = new Simple("IADD", 79);

    /** Defines the IALOAD instruction (no parameter): load int from array. */
    public static final Instruction IALOAD = new Simple("IALOAD", 80);

    /** Defines the IAND instruction (no parameter): boolean AND int. */
    public static final Instruction IAND = new Simple("IAND", 81);

    /** Defines the IASTORE instruction (no parameter): store into int array. */
    public static final Instruction IASTORE = new Simple("IASTORE", 82);

    /** Defines the ICONST_0 instruction (no parameter): push int constant. */
    public static final Instruction ICONST_0 = new Simple("ICONST_0", 83);

    /** Defines the ICONST_1 instruction (no parameter): push int constant. */
    public static final Instruction ICONST_1 = new Simple("ICONST_1", 84);

    /** Defines the ICONST_2 instruction (no parameter): push int constant. */
    public static final Instruction ICONST_2 = new Simple("ICONST_2", 85);

    /** Defines the ICONST_3 instruction (no parameter): push int constant. */
    public static final Instruction ICONST_3 = new Simple("ICONST_3", 86);

    /** Defines the ICONST_4 instruction (no parameter): push int constant. */
    public static final Instruction ICONST_4 = new Simple("ICONST_4", 87);

    /** Defines the ICONST_5 instruction (no parameter): push int constant. */
    public static final Instruction ICONST_5 = new Simple("ICONST_5", 88);

    /** Defines the ICONST_M1 instruction (no parameter): push int constant. */
    public static final Instruction ICONST_M1 = new Simple("ICONST_M1", 89);

    /** Defines the IDIV instruction (no parameter): divide int. */
    public static final Instruction IDIV = new Simple("IDIV", 90);

    /** Defines the ILOAD_0 instruction (no parameter): load int from local variable. */
    public static final Instruction ILOAD_0 = new Simple("ILOAD_0", 91);

    /** Defines the ILOAD_1 instruction (no parameter): load int from local variable. */
    public static final Instruction ILOAD_1 = new Simple("ILOAD_1", 92);

    /** Defines the ILOAD_2 instruction (no parameter): load int from local variable. */
    public static final Instruction ILOAD_2 = new Simple("ILOAD_2", 93);

    /** Defines the ILOAD_3 instruction (no parameter): load int from local variable. */
    public static final Instruction ILOAD_3 = new Simple("ILOAD_3", 94);

    /** Defines the IMUL instruction (no parameter): multiply int. */
    public static final Instruction IMUL = new Simple("IMUL", 95);

    /** Defines the INEG instruction (no parameter): negate int. */
    public static final Instruction INEG = new Simple("INEG", 96);

    /** Defines the IOR instruction (no parameter): boolean OR int. */
    public static final Instruction IOR = new Simple("IOR", 97);

    /** Defines the IREM instruction (no parameter): remainder int. */
    public static final Instruction IREM = new Simple("IREM", 98);

    /** Defines the IRETURN instruction (no parameter): return int from method. */
    public static final Instruction IRETURN = new Simple("IRETURN", 99);

    /** Defines the ISHL instruction (no parameter): shift left int. */
    public static final Instruction ISHL = new Simple("ISHL", 100);

    /** Defines the ISHR instruction (no parameter): arithmetic shift right int. */
    public static final Instruction ISHR = new Simple("ISHR", 101);

    /** Defines the ISTORE_0 instruction (no parameter): store int into local variable. */
    public static final Instruction ISTORE_0 = new Simple("ISTORE_0", 102);

    /** Defines the ISTORE_1 instruction (no parameter): store int into local variable. */
    public static final Instruction ISTORE_1 = new Simple("ISTORE_1", 103);

    /** Defines the ISTORE_2 instruction (no parameter): store int into local variable. */
    public static final Instruction ISTORE_2 = new Simple("ISTORE_2", 104);

    /** Defines the ISTORE_3 instruction (no parameter): store int into local variable. */
    public static final Instruction ISTORE_3 = new Simple("ISTORE_3", 105);

    /** Defines the ISUB instruction (no parameter): subtract int. */
    public static final Instruction ISUB = new Simple("ISUB", 106);

    /** Defines the IUSHR instruction (no parameter): logical shift right int. */
    public static final Instruction IUSHR = new Simple("IUSHR", 107);

    /** Defines the IXOR instruction (no parameter): boolean XOR int. */
    public static final Instruction IXOR = new Simple("IXOR", 108);

    /** Defines the L2D instruction (no parameter): convert long to double. */
    public static final Instruction L2D = new Simple("L2D", 109);

    /** Defines the L2F instruction (no parameter): convert long to float. */
    public static final Instruction L2F = new Simple("L2F", 110);

    /** Defines the L2I instruction (no parameter): convert long to int. */
    public static final Instruction L2I = new Simple("L2I", 111);

    /** Defines the LADD instruction (no parameter): add long. */
    public static final Instruction LADD = new Simple("LADD", 112);

    /** Defines the LALOAD instruction (no parameter): load long from array. */
    public static final Instruction LALOAD = new Simple("LALOAD", 113);

    /** Defines the LAND instruction (no parameter): boolean AND long. */
    public static final Instruction LAND = new Simple("LAND", 114);

    /** Defines the LASTORE instruction (no parameter): store into long array. */
    public static final Instruction LASTORE = new Simple("LASTORE", 115);

    /** Defines the LCMP instruction (no parameter): compare long. */
    public static final Instruction LCMP = new Simple("LCMP", 116);

    /** Defines the LCONST_0 instruction (no parameter): push long constant. */
    public static final Instruction LCONST_0 = new Simple("LCONST_0", 117);

    /** Defines the LCONST_1 instruction (no parameter): push long constant. */
    public static final Instruction LCONST_1 = new Simple("LCONST_1", 118);

    /** Defines the LDIV instruction (no parameter): divide long. */
    public static final Instruction LDIV = new Simple("LDIV", 119);

    /** Defines the LLOAD_0 instruction (no parameter): load long from local variable. */
    public static final Instruction LLOAD_0 = new Simple("LLOAD_0", 120);

    /** Defines the LLOAD_1 instruction (no parameter): load long from local variable. */
    public static final Instruction LLOAD_1 = new Simple("LLOAD_1", 121);

    /** Defines the LLOAD_2 instruction (no parameter): load long from local variable. */
    public static final Instruction LLOAD_2 = new Simple("LLOAD_2", 122);

    /** Defines the LLOAD_3 instruction (no parameter): load long from local variable. */
    public static final Instruction LLOAD_3 = new Simple("LLOAD_3", 123);

    /** Defines the LMUL instruction (no parameter): multiply long. */
    public static final Instruction LMUL = new Simple("LMUL", 124);

    /** Defines the LNEG instruction (no parameter): negate long. */
    public static final Instruction LNEG = new Simple("LNEG", 125);

    /** Defines the LOR instruction (no parameter): boolean OR long. */
    public static final Instruction LOR = new Simple("LOR", 126);

    /** Defines the LREM instruction (no parameter): remainder long. */
    public static final Instruction LREM = new Simple("LREM", 127);

    /** Defines the LRETURN instruction (no parameter): return long from method. */
    public static final Instruction LRETURN = new Simple("LRETURN", 128);

    /** Defines the LSHL instruction (no parameter): shift left long. */
    public static final Instruction LSHL = new Simple("LSHL", 129);

    /** Defines the LSHR instruction (no parameter): arithmetic shift right long. */
    public static final Instruction LSHR = new Simple("LSHR", 130);

    /** Defines the LSTORE_0 instruction (no parameter): store long into local variable. */
    public static final Instruction LSTORE_0 = new Simple("LSTORE_0", 131);

    /** Defines the LSTORE_1 instruction (no parameter): store long into local variable. */
    public static final Instruction LSTORE_1 = new Simple("LSTORE_1", 132);

    /** Defines the LSTORE_2 instruction (no parameter): store long into local variable. */
    public static final Instruction LSTORE_2 = new Simple("LSTORE_2", 133);

    /** Defines the LSTORE_3 instruction (no parameter): store long into local variable. */
    public static final Instruction LSTORE_3 = new Simple("LSTORE_3", 134);

    /** Defines the LSUB instruction (no parameter): subtract long. */
    public static final Instruction LSUB = new Simple("LSUB", 135);

    /** Defines the LUSHR instruction (no parameter): logical shift right long. */
    public static final Instruction LUSHR = new Simple("LUSHR", 136);

    /** Defines the LXOR instruction (no parameter): boolean XOR long. */
    public static final Instruction LXOR = new Simple("LXOR", 137);

    /** Defines the MONITORENTER instruction (no parameter): enter monitor for object. */
    public static final Instruction MONITORENTER = new Simple("MONITORENTER", 138);

    /** Defines the MONITOREXIT instruction (no parameter): exit monitor for object. */
    public static final Instruction MONITOREXIT = new Simple("MONITOREXIT", 139);

    /** Defines the NOP instruction (no parameter): do nothing. */
    public static final Instruction NOP = new Simple("NOP", 140);

    /** Defines the POP instruction (no parameter): pop the top operand stack value. */
    public static final Instruction POP = new Simple("POP", 141);

    /** Defines the POP2 instruction (no parameter): pop the top one or two operand stack values. */
    public static final Instruction POP2 = new Simple("POP2", 142);

    /** Defines the RETURN instruction (no parameter): return void from method. */
    public static final Instruction RETURN = new Simple("RETURN", 143);

    /** Defines the SALOAD instruction (no parameter): load short from array. */
    public static final Instruction SALOAD = new Simple("SALOAD", 144);

    /** Defines the SASTORE instruction (no parameter): store into short array. */
    public static final Instruction SASTORE = new Simple("SASTORE", 145);

    /** Defines the SWAP instruction (no parameter): swap the top two operand stack values. */
    public static final Instruction SWAP = new Simple("SWAP", 146);

    /** Defines the ALOAD instruction: load reference from local variable. */
    public static final class ALOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the ALOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public ALOAD(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 0);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ALOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 0;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ALOAD) {
                final ALOAD that = (ALOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ALOAD'

    /** Defines the ANEWARRAY instruction: create new array of references. */
    public static final class ANEWARRAY_TYPE extends Instruction {

        /** Value of first instruction parameter. */
        private final TypeName p1;

        /**
         * Constructs an instance of the ANEWARRAY instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public ANEWARRAY_TYPE(final TypeName p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (TypeName)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public TypeName getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 1);
            b.set(0, wrap("Class_or_interface", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ANEWARRAY_TYPE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 1;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ANEWARRAY_TYPE) {
                final ANEWARRAY_TYPE that = (ANEWARRAY_TYPE) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ANEWARRAY_TYPE'

    /** Defines the ANEWARRAY instruction: create new array of references. */
    public static final class ANEWARRAY_ARRAY extends Instruction {

        /** Value of first instruction parameter. */
        private final Descriptor.Array p1;

        /**
         * Constructs an instance of the ANEWARRAY instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public ANEWARRAY_ARRAY(final Descriptor.Array p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (Descriptor.Array)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public Descriptor.Array getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 1);
            b.set(0, wrap("Array_type", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ANEWARRAY_ARRAY" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 1;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ANEWARRAY_ARRAY) {
                final ANEWARRAY_ARRAY that = (ANEWARRAY_ARRAY) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ANEWARRAY_ARRAY'

    /** Defines the ASTORE instruction: store reference into local variable. */
    public static final class ASTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the ASTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public ASTORE(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 2);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ASTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 2;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ASTORE) {
                final ASTORE that = (ASTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ASTORE'

    /** Defines the BIPUSH instruction: push byte. */
    public static final class BIPUSH extends Instruction {

        /** Value of first instruction parameter. */
        private final byte p1;

        /**
         * Constructs an instance of the BIPUSH instruction.
         * @param p1 value of first instruction parameter
         */
        public BIPUSH(final byte p1) {
            this.p1 = p1;
        } // end constructor (byte)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public byte getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 3);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "BIPUSH" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 3;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof BIPUSH) {
                final BIPUSH that = (BIPUSH) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'BIPUSH'

    /** Defines the CHECKCAST instruction: check whether object is of given type. */
    public static final class CHECKCAST_TYPE extends Instruction {

        /** Value of first instruction parameter. */
        private final TypeName p1;

        /**
         * Constructs an instance of the CHECKCAST instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public CHECKCAST_TYPE(final TypeName p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (TypeName)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public TypeName getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 4);
            b.set(0, wrap("Class_or_interface", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "CHECKCAST_TYPE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 4;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof CHECKCAST_TYPE) {
                final CHECKCAST_TYPE that = (CHECKCAST_TYPE) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'CHECKCAST_TYPE'

    /** Defines the CHECKCAST instruction: check whether object is of given type. */
    public static final class CHECKCAST_ARRAY extends Instruction {

        /** Value of first instruction parameter. */
        private final Descriptor.Array p1;

        /**
         * Constructs an instance of the CHECKCAST instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public CHECKCAST_ARRAY(final Descriptor.Array p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (Descriptor.Array)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public Descriptor.Array getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 4);
            b.set(0, wrap("Array_type", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "CHECKCAST_ARRAY" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 4;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof CHECKCAST_ARRAY) {
                final CHECKCAST_ARRAY that = (CHECKCAST_ARRAY) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'CHECKCAST_ARRAY'

    /** Defines the DLOAD instruction: load double from local variable. */
    public static final class DLOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the DLOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public DLOAD(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 5);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "DLOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 5;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof DLOAD) {
                final DLOAD that = (DLOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'DLOAD'

    /** Defines the DSTORE instruction: store double into local variable. */
    public static final class DSTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the DSTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public DSTORE(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 6);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "DSTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 6;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof DSTORE) {
                final DSTORE that = (DSTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'DSTORE'

    /** Defines the FLOAD instruction: load float from local variable. */
    public static final class FLOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the FLOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public FLOAD(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 7);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "FLOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 7;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof FLOAD) {
                final FLOAD that = (FLOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'FLOAD'

    /** Defines the FSTORE instruction: store float into local variable. */
    public static final class FSTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the FSTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public FSTORE(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 8);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "FSTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 8;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof FSTORE) {
                final FSTORE that = (FSTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'FSTORE'

    /** Defines the GETFIELD instruction: fetch field from object. */
    public static final class GETFIELD extends Instruction {

        /** Value of first instruction parameter. */
        private final FieldRef p1;

        /**
         * Constructs an instance of the GETFIELD instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public GETFIELD(final FieldRef p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (FieldRef)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public FieldRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 9);
            b.set(0, this.p1.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "GETFIELD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 9;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof GETFIELD) {
                final GETFIELD that = (GETFIELD) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'GETFIELD'

    /** Defines the GETSTATIC instruction: get static field from class. */
    public static final class GETSTATIC extends Instruction {

        /** Value of first instruction parameter. */
        private final FieldRef p1;

        /**
         * Constructs an instance of the GETSTATIC instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public GETSTATIC(final FieldRef p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (FieldRef)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public FieldRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 10);
            b.set(0, this.p1.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "GETSTATIC" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 10;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof GETSTATIC) {
                final GETSTATIC that = (GETSTATIC) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'GETSTATIC'

    /** Defines the GOTO instruction: branch always. */
    public static final class GOTO extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the GOTO instruction.
         * @param p1 value of first instruction parameter
         */
        public GOTO(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 11);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "GOTO" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 11;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof GOTO) {
                final GOTO that = (GOTO) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'GOTO'

    /** Defines the GOTO_W instruction: branch always. */
    public static final class GOTO_W extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the GOTO_W instruction.
         * @param p1 value of first instruction parameter
         */
        public GOTO_W(final int p1) {
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 12);
            b.set(0, Helper.createInt32(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "GOTO_W" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 12;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof GOTO_W) {
                final GOTO_W that = (GOTO_W) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'GOTO_W'

    /** Defines the IF_ACMPEQ instruction: branch if reference comparison succeeds. */
    public static final class IF_ACMPEQ extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IF_ACMPEQ instruction.
         * @param p1 value of first instruction parameter
         */
        public IF_ACMPEQ(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 13);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IF_ACMPEQ" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 13;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IF_ACMPEQ) {
                final IF_ACMPEQ that = (IF_ACMPEQ) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IF_ACMPEQ'

    /** Defines the IF_ACMPNE instruction: branch if reference comparison succeeds. */
    public static final class IF_ACMPNE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IF_ACMPNE instruction.
         * @param p1 value of first instruction parameter
         */
        public IF_ACMPNE(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 14);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IF_ACMPNE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 14;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IF_ACMPNE) {
                final IF_ACMPNE that = (IF_ACMPNE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IF_ACMPNE'

    /** Defines the IF_ICMPEQ instruction: branch if int comparison succeeds. */
    public static final class IF_ICMPEQ extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IF_ICMPEQ instruction.
         * @param p1 value of first instruction parameter
         */
        public IF_ICMPEQ(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 15);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IF_ICMPEQ" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 15;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IF_ICMPEQ) {
                final IF_ICMPEQ that = (IF_ICMPEQ) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IF_ICMPEQ'

    /** Defines the IF_ICMPGE instruction: branch if int comparison succeeds. */
    public static final class IF_ICMPGE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IF_ICMPGE instruction.
         * @param p1 value of first instruction parameter
         */
        public IF_ICMPGE(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 16);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IF_ICMPGE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 16;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IF_ICMPGE) {
                final IF_ICMPGE that = (IF_ICMPGE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IF_ICMPGE'

    /** Defines the IF_ICMPGT instruction: branch if int comparison succeeds. */
    public static final class IF_ICMPGT extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IF_ICMPGT instruction.
         * @param p1 value of first instruction parameter
         */
        public IF_ICMPGT(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 17);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IF_ICMPGT" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 17;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IF_ICMPGT) {
                final IF_ICMPGT that = (IF_ICMPGT) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IF_ICMPGT'

    /** Defines the IF_ICMPLE instruction: branch if int comparison succeeds. */
    public static final class IF_ICMPLE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IF_ICMPLE instruction.
         * @param p1 value of first instruction parameter
         */
        public IF_ICMPLE(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 18);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IF_ICMPLE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 18;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IF_ICMPLE) {
                final IF_ICMPLE that = (IF_ICMPLE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IF_ICMPLE'

    /** Defines the IF_ICMPLT instruction: branch if int comparison succeeds. */
    public static final class IF_ICMPLT extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IF_ICMPLT instruction.
         * @param p1 value of first instruction parameter
         */
        public IF_ICMPLT(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 19);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IF_ICMPLT" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 19;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IF_ICMPLT) {
                final IF_ICMPLT that = (IF_ICMPLT) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IF_ICMPLT'

    /** Defines the IF_ICMPNE instruction: branch if int comparison succeeds. */
    public static final class IF_ICMPNE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IF_ICMPNE instruction.
         * @param p1 value of first instruction parameter
         */
        public IF_ICMPNE(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 20);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IF_ICMPNE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 20;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IF_ICMPNE) {
                final IF_ICMPNE that = (IF_ICMPNE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IF_ICMPNE'

    /** Defines the IFEQ instruction: branch if int comparison with zero succeeds. */
    public static final class IFEQ extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IFEQ instruction.
         * @param p1 value of first instruction parameter
         */
        public IFEQ(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 21);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IFEQ" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 21;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IFEQ) {
                final IFEQ that = (IFEQ) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IFEQ'

    /** Defines the IFGE instruction: branch if int comparison with zero succeeds. */
    public static final class IFGE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IFGE instruction.
         * @param p1 value of first instruction parameter
         */
        public IFGE(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 22);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IFGE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 22;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IFGE) {
                final IFGE that = (IFGE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IFGE'

    /** Defines the IFGT instruction: branch if int comparison with zero succeeds. */
    public static final class IFGT extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IFGT instruction.
         * @param p1 value of first instruction parameter
         */
        public IFGT(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 23);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IFGT" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 23;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IFGT) {
                final IFGT that = (IFGT) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IFGT'

    /** Defines the IFLE instruction: branch if int comparison with zero succeeds. */
    public static final class IFLE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IFLE instruction.
         * @param p1 value of first instruction parameter
         */
        public IFLE(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 24);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IFLE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 24;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IFLE) {
                final IFLE that = (IFLE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IFLE'

    /** Defines the IFLT instruction: branch if int comparison with zero succeeds. */
    public static final class IFLT extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IFLT instruction.
         * @param p1 value of first instruction parameter
         */
        public IFLT(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 25);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IFLT" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 25;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IFLT) {
                final IFLT that = (IFLT) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IFLT'

    /** Defines the IFNE instruction: branch if int comparison with zero succeeds. */
    public static final class IFNE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IFNE instruction.
         * @param p1 value of first instruction parameter
         */
        public IFNE(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 26);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IFNE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 26;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IFNE) {
                final IFNE that = (IFNE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IFNE'

    /** Defines the IFNONNULL instruction: branch if reference not null. */
    public static final class IFNONNULL extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IFNONNULL instruction.
         * @param p1 value of first instruction parameter
         */
        public IFNONNULL(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 27);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IFNONNULL" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 27;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IFNONNULL) {
                final IFNONNULL that = (IFNONNULL) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IFNONNULL'

    /** Defines the IFNULL instruction: branch if reference not null. */
    public static final class IFNULL extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the IFNULL instruction.
         * @param p1 value of first instruction parameter
         */
        public IFNULL(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 28);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IFNULL" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 28;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IFNULL) {
                final IFNULL that = (IFNULL) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IFNULL'

    /** Defines the IINC instruction: increment local variable by constant. */
    public static final class IINC extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /** Value of second instruction parameter. */
        private final byte p2;

        /**
         * Constructs an instance of the IINC instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         * @param p2 value of second instruction parameter
         */
        public IINC(final short p1, final byte p2) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
            this.p2 = p2;
        } // end constructor (short, byte)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * Returns the value of the second parameter.
         * @return the value of the second parameter
         */
        public byte getSecondParameter() {
            return this.p2;
        } // end method 'getSecondParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(2, 29);
            b.set(0, Helper.createInt(this.p1));
            b.set(1, Helper.createInt(this.p2));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "IINC" + " " + this.p1 + " " + this.p2;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 29;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof IINC) {
                final IINC that = (IINC) obj;
                return this.p1 == that.p1
                    && this.p2 == that.p2;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IINC'

    /** Defines the ILOAD instruction: load int from local variable. */
    public static final class ILOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the ILOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public ILOAD(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 30);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ILOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 30;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ILOAD) {
                final ILOAD that = (ILOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ILOAD'

    /** Defines the INSTANCEOF instruction: determine if object is of given type. */
    public static final class INSTANCEOF_TYPE extends Instruction {

        /** Value of first instruction parameter. */
        private final TypeName p1;

        /**
         * Constructs an instance of the INSTANCEOF instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public INSTANCEOF_TYPE(final TypeName p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (TypeName)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public TypeName getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 31);
            b.set(0, wrap("Class_or_interface", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "INSTANCEOF_TYPE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 31;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof INSTANCEOF_TYPE) {
                final INSTANCEOF_TYPE that = (INSTANCEOF_TYPE) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'INSTANCEOF_TYPE'

    /** Defines the INSTANCEOF instruction: determine if object is of given type. */
    public static final class INSTANCEOF_ARRAY extends Instruction {

        /** Value of first instruction parameter. */
        private final Descriptor.Array p1;

        /**
         * Constructs an instance of the INSTANCEOF instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public INSTANCEOF_ARRAY(final Descriptor.Array p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (Descriptor.Array)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public Descriptor.Array getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 31);
            b.set(0, wrap("Array_type", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "INSTANCEOF_ARRAY" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 31;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof INSTANCEOF_ARRAY) {
                final INSTANCEOF_ARRAY that = (INSTANCEOF_ARRAY) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'INSTANCEOF_ARRAY'

    /** Defines the INVOKEDYNAMIC instruction: invoke instance method; resolve and dispatch based on class. */
    public static final class INVOKEDYNAMIC extends Instruction {

        /** Value of first instruction parameter. */
        private final DynamicMethodRef p1;

        /**
         * Constructs an instance of the INVOKEDYNAMIC instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public INVOKEDYNAMIC(final DynamicMethodRef p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (DynamicMethodRef)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public DynamicMethodRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 32);
            b.set(0, this.p1.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "INVOKEDYNAMIC" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 32;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof INVOKEDYNAMIC) {
                final INVOKEDYNAMIC that = (INVOKEDYNAMIC) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'INVOKEDYNAMIC'

    /** Defines the INVOKEINTERFACE instruction: invoke interface method. */
    public static final class INVOKEINTERFACE extends Instruction {

        /** Value of first instruction parameter. */
        private final MethodRef p1;

        /** Value of second instruction parameter. */
        private final short p2;

        /**
         * Constructs an instance of the INVOKEINTERFACE instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         * @param p2 value of second instruction parameter - should be in 0..255
         */
        public INVOKEINTERFACE(final MethodRef p1, final short p2) {
            assert p1 != null : "null value";
            assert p2 >= 0 && p2 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
            this.p2 = p2;
        } // end constructor (MethodRef, short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public MethodRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * Returns the value of the second parameter.
         * @return the value of the second parameter
         */
        public short getSecondParameter() {
            return this.p2;
        } // end method 'getSecondParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(2, 33);
            b.set(0, this.p1.toValue());
            b.set(1, Helper.createInt(this.p2));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "INVOKEINTERFACE" + " " + this.p1 + " " + this.p2;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 33;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof INVOKEINTERFACE) {
                final INVOKEINTERFACE that = (INVOKEINTERFACE) obj;
                return this.p1.equals(that.p1)
                    && this.p2 == that.p2;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'INVOKEINTERFACE'

    /** Defines the INVOKESPECIAL instruction: invoke instance method; special handling for superclass, private, and instance initialization method invocations. */
    public static final class INVOKESPECIAL extends Instruction {

        /** Value of first instruction parameter. */
        private final MethodRef p1;

        /**
         * Constructs an instance of the INVOKESPECIAL instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public INVOKESPECIAL(final MethodRef p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (MethodRef)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public MethodRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 34);
            b.set(0, this.p1.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "INVOKESPECIAL" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 34;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof INVOKESPECIAL) {
                final INVOKESPECIAL that = (INVOKESPECIAL) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'INVOKESPECIAL'

    /** Defines the INVOKESTATIC instruction: invoke a class (static) method. */
    public static final class INVOKESTATIC extends Instruction {

        /** Value of first instruction parameter. */
        private final MethodRef p1;

        /**
         * Constructs an instance of the INVOKESTATIC instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public INVOKESTATIC(final MethodRef p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (MethodRef)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public MethodRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 35);
            b.set(0, this.p1.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "INVOKESTATIC" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 35;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof INVOKESTATIC) {
                final INVOKESTATIC that = (INVOKESTATIC) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'INVOKESTATIC'

    /** Defines the INVOKEVIRTUAL instruction: invoke instance method; dispatch based on class. */
    public static final class INVOKEVIRTUAL extends Instruction {

        /** Value of first instruction parameter. */
        private final MethodRef p1;

        /**
         * Constructs an instance of the INVOKEVIRTUAL instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public INVOKEVIRTUAL(final MethodRef p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (MethodRef)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public MethodRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 36);
            b.set(0, this.p1.toTaggedValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "INVOKEVIRTUAL" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 36;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof INVOKEVIRTUAL) {
                final INVOKEVIRTUAL that = (INVOKEVIRTUAL) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'INVOKEVIRTUAL'

    /** Defines the INVOKEVIRTUAL instruction: invoke instance method; dispatch based on class. */
    public static final class INVOKEVIRTUAL_ARRAY extends Instruction {

        /** Value of first instruction parameter. */
        private final ArrayMethodRef p1;

        /**
         * Constructs an instance of the INVOKEVIRTUAL instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public INVOKEVIRTUAL_ARRAY(final ArrayMethodRef p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (ArrayMethodRef)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public ArrayMethodRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 36);
            b.set(0, this.p1.toTaggedValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "INVOKEVIRTUAL_ARRAY" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 36;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof INVOKEVIRTUAL_ARRAY) {
                final INVOKEVIRTUAL_ARRAY that = (INVOKEVIRTUAL_ARRAY) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'INVOKEVIRTUAL_ARRAY'

    /** Defines the ISTORE instruction: store int into local variable. */
    public static final class ISTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the ISTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public ISTORE(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 37);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ISTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 37;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof ISTORE) {
                final ISTORE that = (ISTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ISTORE'

    /** Defines the JSR instruction: jump subroutine. */
    public static final class JSR extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the JSR instruction.
         * @param p1 value of first instruction parameter
         */
        public JSR(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 38);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "JSR" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 38;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof JSR) {
                final JSR that = (JSR) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'JSR'

    /** Defines the JSR_W instruction: jump subroutine (wide index). */
    public static final class JSR_W extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the JSR_W instruction.
         * @param p1 value of first instruction parameter
         */
        public JSR_W(final int p1) {
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 39);
            b.set(0, Helper.createInt32(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "JSR_W" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 39;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof JSR_W) {
                final JSR_W that = (JSR_W) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'JSR_W'

    /** Defines the LDC instruction: push item from runtime constant pool. */
    public static final class LDC_INT extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the LDC instruction.
         * @param p1 value of first instruction parameter
         */
        public LDC_INT(final int p1) {
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 40);
            b.set(0, wrap("Int", Helper.createInt32(this.p1)));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_INT" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 40;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_INT) {
                final LDC_INT that = (LDC_INT) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_INT'

    /** Defines the LDC instruction: push item from runtime constant pool. */
    public static final class LDC_FLOAT extends Instruction {

        /** Value of first instruction parameter. */
        private final float p1;

        /**
         * Constructs an instance of the LDC instruction.
         * @param p1 value of first instruction parameter
         */
        public LDC_FLOAT(final float p1) {
            this.p1 = p1;
        } // end constructor (float)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public float getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 40);
            b.set(0, wrap("Float", Helper.createFloat(this.p1)));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_FLOAT" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 40;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_FLOAT) {
                final LDC_FLOAT that = (LDC_FLOAT) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_FLOAT'

    /** Defines the LDC instruction: push item from runtime constant pool. */
    public static final class LDC_STRING extends Instruction {

        /** Value of first instruction parameter. */
        private final String p1;

        /**
         * Constructs an instance of the LDC instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public LDC_STRING(final String p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (String)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public String getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 40);
            b.set(0, wrap("String", Helper.createString(this.p1)));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_STRING" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 40;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_STRING) {
                final LDC_STRING that = (LDC_STRING) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_STRING'

    /** Defines the LDC instruction: push item from runtime constant pool. */
    public static final class LDC_TYPE extends Instruction {

        /** Value of first instruction parameter. */
        private final TypeName p1;

        /**
         * Constructs an instance of the LDC instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public LDC_TYPE(final TypeName p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (TypeName)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public TypeName getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 40);
            b.set(0, wrap("Class_or_interface", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_TYPE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 40;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_TYPE) {
                final LDC_TYPE that = (LDC_TYPE) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_TYPE'

    /** Defines the LDC instruction: push item from runtime constant pool. */
    public static final class LDC_ARRAY extends Instruction {

        /** Value of first instruction parameter. */
        private final Descriptor.Array p1;

        /**
         * Constructs an instance of the LDC instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public LDC_ARRAY(final Descriptor.Array p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (Descriptor.Array)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public Descriptor.Array getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 40);
            b.set(0, wrap("Array_type", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_ARRAY" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 40;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_ARRAY) {
                final LDC_ARRAY that = (LDC_ARRAY) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_ARRAY'

    /** Defines the LDC2_W instruction: push long or double from runtime constant pool (wide index). */
    public static final class LDC2_W_LONG extends Instruction {

        /** Value of first instruction parameter. */
        private final long p1;

        /**
         * Constructs an instance of the LDC2_W instruction.
         * @param p1 value of first instruction parameter
         */
        public LDC2_W_LONG(final long p1) {
            this.p1 = p1;
        } // end constructor (long)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public long getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 41);
            b.set(0, wrap("Long", Helper.createInt64(this.p1)));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC2_W_LONG" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 41;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC2_W_LONG) {
                final LDC2_W_LONG that = (LDC2_W_LONG) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC2_W_LONG'

    /** Defines the LDC2_W instruction: push long or double from runtime constant pool (wide index). */
    public static final class LDC2_W_DOUBLE extends Instruction {

        /** Value of first instruction parameter. */
        private final double p1;

        /**
         * Constructs an instance of the LDC2_W instruction.
         * @param p1 value of first instruction parameter
         */
        public LDC2_W_DOUBLE(final double p1) {
            this.p1 = p1;
        } // end constructor (double)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public double getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 41);
            b.set(0, wrap("Double", Helper.createFloat(this.p1)));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC2_W_DOUBLE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 41;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC2_W_DOUBLE) {
                final LDC2_W_DOUBLE that = (LDC2_W_DOUBLE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC2_W_DOUBLE'

    /** Defines the LDC_W instruction: push item from runtime constant pool (wide index). */
    public static final class LDC_W_INT extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the LDC_W instruction.
         * @param p1 value of first instruction parameter
         */
        public LDC_W_INT(final int p1) {
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 42);
            b.set(0, wrap("Int", Helper.createInt32(this.p1)));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_W_INT" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 42;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_W_INT) {
                final LDC_W_INT that = (LDC_W_INT) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_W_INT'

    /** Defines the LDC_W instruction: push item from runtime constant pool (wide index). */
    public static final class LDC_W_FLOAT extends Instruction {

        /** Value of first instruction parameter. */
        private final float p1;

        /**
         * Constructs an instance of the LDC_W instruction.
         * @param p1 value of first instruction parameter
         */
        public LDC_W_FLOAT(final float p1) {
            this.p1 = p1;
        } // end constructor (float)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public float getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 42);
            b.set(0, wrap("Float", Helper.createFloat(this.p1)));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_W_FLOAT" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 42;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_W_FLOAT) {
                final LDC_W_FLOAT that = (LDC_W_FLOAT) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_W_FLOAT'

    /** Defines the LDC_W instruction: push item from runtime constant pool (wide index). */
    public static final class LDC_W_STRING extends Instruction {

        /** Value of first instruction parameter. */
        private final String p1;

        /**
         * Constructs an instance of the LDC_W instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public LDC_W_STRING(final String p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (String)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public String getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 42);
            b.set(0, wrap("String", Helper.createString(this.p1)));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_W_STRING" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 42;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_W_STRING) {
                final LDC_W_STRING that = (LDC_W_STRING) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_W_STRING'

    /** Defines the LDC_W instruction: push item from runtime constant pool (wide index). */
    public static final class LDC_W_TYPE extends Instruction {

        /** Value of first instruction parameter. */
        private final TypeName p1;

        /**
         * Constructs an instance of the LDC_W instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public LDC_W_TYPE(final TypeName p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (TypeName)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public TypeName getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 42);
            b.set(0, wrap("Class_or_interface", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_W_TYPE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 42;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_W_TYPE) {
                final LDC_W_TYPE that = (LDC_W_TYPE) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_W_TYPE'

    /** Defines the LDC_W instruction: push item from runtime constant pool (wide index). */
    public static final class LDC_W_ARRAY extends Instruction {

        /** Value of first instruction parameter. */
        private final Descriptor.Array p1;

        /**
         * Constructs an instance of the LDC_W instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public LDC_W_ARRAY(final Descriptor.Array p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (Descriptor.Array)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public Descriptor.Array getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 42);
            b.set(0, wrap("Array_type", this.p1.toValue()));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LDC_W_ARRAY" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 42;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LDC_W_ARRAY) {
                final LDC_W_ARRAY that = (LDC_W_ARRAY) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LDC_W_ARRAY'

    /** Defines the LLOAD instruction: load long from local variable. */
    public static final class LLOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the LLOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public LLOAD(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 43);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LLOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 43;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LLOAD) {
                final LLOAD that = (LLOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LLOAD'

    /** Defines the LOOKUPSWITCH instruction: access jump table by key match and jump. */
    public static final class LOOKUPSWITCH extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /** Value of second instruction parameter. */
        private final int p2;

        /** Value of third instruction parameter. */
        private final List<MatchOffsetPair> p3;

        /**
         * Constructs an instance of the LOOKUPSWITCH instruction.
         * @param p1 value of first instruction parameter
         * @param p2 value of second instruction parameter
         * @param p3 value of third instruction parameter - should not be <tt>null</tt>
         */
        public LOOKUPSWITCH(final int p1, final int p2, final List<MatchOffsetPair> p3) {
            assert p3 != null : "null value";
            this.p1 = p1;
            this.p2 = p2;
            this.p3 = new LinkedList<MatchOffsetPair>(p3);
        } // end constructor (int, int, List<MatchOffsetPair>)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * Returns the value of the second parameter.
         * @return the value of the second parameter
         */
        public int getSecondParameter() {
            return this.p2;
        } // end method 'getSecondParameter()'

        /**
         * Returns the value of the third parameter.
         * @return the value of the third parameter
         */
        public List<MatchOffsetPair> getThirdParameter() {
            return new LinkedList<MatchOffsetPair>(this.p3);
        } // end method 'getThirdParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(3, 44);
            b.set(0, Helper.createInt32(this.p1));
            b.set(1, Helper.createInt32(this.p2));
            b.set(2, API.encodeList(this.p3));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LOOKUPSWITCH" + " " + this.p1 + " " + this.p2 + " " + this.p3;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 44;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LOOKUPSWITCH) {
                final LOOKUPSWITCH that = (LOOKUPSWITCH) obj;
                return this.p1 == that.p1
                    && this.p2 == that.p2
                    && this.p3.equals(that.p3);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LOOKUPSWITCH'

    /** Defines the LSTORE instruction: store long into local variable. */
    public static final class LSTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the LSTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public LSTORE(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 45);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "LSTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 45;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof LSTORE) {
                final LSTORE that = (LSTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LSTORE'

    /** Defines the MULTIANEWARRAY instruction: create new multidimensional array. */
    public static final class MULTIANEWARRAY_TYPE extends Instruction {

        /** Value of first instruction parameter. */
        private final TypeName p1;

        /** Value of second instruction parameter. */
        private final short p2;

        /**
         * Constructs an instance of the MULTIANEWARRAY instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         * @param p2 value of second instruction parameter - should be in 0..255
         */
        public MULTIANEWARRAY_TYPE(final TypeName p1, final short p2) {
            assert p1 != null : "null value";
            assert p2 >= 0 && p2 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
            this.p2 = p2;
        } // end constructor (TypeName, short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public TypeName getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * Returns the value of the second parameter.
         * @return the value of the second parameter
         */
        public short getSecondParameter() {
            return this.p2;
        } // end method 'getSecondParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(2, 46);
            b.set(0, wrap("Class_or_interface", this.p1.toValue()));
            b.set(1, Helper.createInt(this.p2));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "MULTIANEWARRAY_TYPE" + " " + this.p1 + " " + this.p2;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 46;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof MULTIANEWARRAY_TYPE) {
                final MULTIANEWARRAY_TYPE that = (MULTIANEWARRAY_TYPE) obj;
                return this.p1.equals(that.p1)
                    && this.p2 == that.p2;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'MULTIANEWARRAY_TYPE'

    /** Defines the MULTIANEWARRAY instruction: create new multidimensional array. */
    public static final class MULTIANEWARRAY_ARRAY extends Instruction {

        /** Value of first instruction parameter. */
        private final Descriptor.Array p1;

        /** Value of second instruction parameter. */
        private final short p2;

        /**
         * Constructs an instance of the MULTIANEWARRAY instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         * @param p2 value of second instruction parameter - should be in 0..255
         */
        public MULTIANEWARRAY_ARRAY(final Descriptor.Array p1, final short p2) {
            assert p1 != null : "null value";
            assert p2 >= 0 && p2 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
            this.p2 = p2;
        } // end constructor (Descriptor.Array, short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public Descriptor.Array getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * Returns the value of the second parameter.
         * @return the value of the second parameter
         */
        public short getSecondParameter() {
            return this.p2;
        } // end method 'getSecondParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(2, 46);
            b.set(0, wrap("Array_type", this.p1.toValue()));
            b.set(1, Helper.createInt(this.p2));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "MULTIANEWARRAY_ARRAY" + " " + this.p1 + " " + this.p2;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 46;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof MULTIANEWARRAY_ARRAY) {
                final MULTIANEWARRAY_ARRAY that = (MULTIANEWARRAY_ARRAY) obj;
                return this.p1.equals(that.p1)
                    && this.p2 == that.p2;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'MULTIANEWARRAY_ARRAY'

    /** Defines the NEW instruction: create new object. */
    public static final class NEW extends Instruction {

        /** Value of first instruction parameter. */
        private final TypeName p1;

        /**
         * Constructs an instance of the NEW instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public NEW(final TypeName p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (TypeName)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public TypeName getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 47);
            b.set(0, this.p1.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "NEW" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 47;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof NEW) {
                final NEW that = (NEW) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'NEW'

    /** Defines the NEWARRAY instruction: create new arrayhandler. */
    public static final class NEWARRAY extends Instruction {

        /** Value of first instruction parameter. */
        private final Descriptor p1;

        /**
         * Constructs an instance of the NEWARRAY instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public NEWARRAY(final Descriptor p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (Descriptor)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public Descriptor getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 48);
            b.set(0, this.p1.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "NEWARRAY" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 48;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof NEWARRAY) {
                final NEWARRAY that = (NEWARRAY) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'NEWARRAY'

    /** Defines the PUTFIELD instruction: set field in object. */
    public static final class PUTFIELD extends Instruction {

        /** Value of first instruction parameter. */
        private final FieldRef p1;

        /**
         * Constructs an instance of the PUTFIELD instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public PUTFIELD(final FieldRef p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (FieldRef)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public FieldRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 49);
            b.set(0, this.p1.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "PUTFIELD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 49;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof PUTFIELD) {
                final PUTFIELD that = (PUTFIELD) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'PUTFIELD'

    /** Defines the PUTSTATIC instruction: set static field in class. */
    public static final class PUTSTATIC extends Instruction {

        /** Value of first instruction parameter. */
        private final FieldRef p1;

        /**
         * Constructs an instance of the PUTSTATIC instruction.
         * @param p1 value of first instruction parameter - should not be <tt>null</tt>
         */
        public PUTSTATIC(final FieldRef p1) {
            assert p1 != null : "null value";
            this.p1 = p1;
        } // end constructor (FieldRef)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public FieldRef getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 50);
            b.set(0, this.p1.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "PUTSTATIC" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 50;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof PUTSTATIC) {
                final PUTSTATIC that = (PUTSTATIC) obj;
                return this.p1.equals(that.p1);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'PUTSTATIC'

    /** Defines the RET instruction: return from subroutine. */
    public static final class RET extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the RET instruction.
         * @param p1 value of first instruction parameter - should be in 0..255
         */
        public RET(final short p1) {
            assert p1 >= 0 && p1 <= 255 : "invalid unsigned byte";
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 51);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "RET" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 51;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof RET) {
                final RET that = (RET) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'RET'

    /** Defines the SIPUSH instruction: push short. */
    public static final class SIPUSH extends Instruction {

        /** Value of first instruction parameter. */
        private final short p1;

        /**
         * Constructs an instance of the SIPUSH instruction.
         * @param p1 value of first instruction parameter
         */
        public SIPUSH(final short p1) {
            this.p1 = p1;
        } // end constructor (short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public short getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 52);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "SIPUSH" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 52;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof SIPUSH) {
                final SIPUSH that = (SIPUSH) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'SIPUSH'

    /** Defines the TABLESWITCH instruction: access jump table by index and jump. */
    public static final class TABLESWITCH extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /** Value of second instruction parameter. */
        private final int p2;

        /** Value of third instruction parameter. */
        private final int p3;

        /** Value of fourth instruction parameter. */
        private final List<Integer> p4;

        /**
         * Constructs an instance of the TABLESWITCH instruction.
         * @param p1 value of first instruction parameter
         * @param p2 value of second instruction parameter
         * @param p3 value of third instruction parameter
         * @param p4 value of fourth instruction parameter - should not be <tt>null</tt>
         */
        public TABLESWITCH(final int p1, final int p2, final int p3, final List<Integer> p4) {
            assert p4 != null : "null value";
            assert (p3 - p2) + 1 != p4.size() : "invalid list length";
            this.p1 = p1;
            this.p2 = p2;
            this.p3 = p3;
            this.p4 = new LinkedList<Integer>(p4);
        } // end constructor (int, int, int, List<Integer>)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * Returns the value of the second parameter.
         * @return the value of the second parameter
         */
        public int getSecondParameter() {
            return this.p2;
        } // end method 'getSecondParameter()'

        /**
         * Returns the value of the third parameter.
         * @return the value of the third parameter
         */
        public int getThirdParameter() {
            return this.p3;
        } // end method 'getThirdParameter()'

        /**
         * Returns the value of the fourth parameter.
         * @return the value of the fourth parameter
         */
        public List<Integer> getFourthParameter() {
            return new LinkedList<Integer>(this.p4);
        } // end method 'getFourthParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(4, 53);
            b.set(0, Helper.createInt32(this.p1));
            b.set(1, Helper.createInt32(this.p2));
            b.set(2, Helper.createInt32(this.p3));
            b.set(3, encodeIntegerList(this.p4));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "TABLESWITCH" + " " + this.p1 + " " + this.p2 + " " + this.p3 + " " + this.p4;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 53;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof TABLESWITCH) {
                final TABLESWITCH that = (TABLESWITCH) obj;
                return this.p1 == that.p1
                    && this.p2 == that.p2
                    && this.p3 == that.p3
                    && this.p4.equals(that.p4);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'TABLESWITCH'

    /** Defines the WIDE_ALOAD instruction: load reference from local variable. */
    public static final class WIDE_ALOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_ALOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_ALOAD(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 54);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_ALOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 54;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_ALOAD) {
                final WIDE_ALOAD that = (WIDE_ALOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_ALOAD'

    /** Defines the WIDE_ASTORE instruction: store reference into local variable. */
    public static final class WIDE_ASTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_ASTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_ASTORE(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 55);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_ASTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 55;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_ASTORE) {
                final WIDE_ASTORE that = (WIDE_ASTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_ASTORE'

    /** Defines the WIDE_DLOAD instruction: load double from local variable. */
    public static final class WIDE_DLOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_DLOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_DLOAD(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 56);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_DLOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 56;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_DLOAD) {
                final WIDE_DLOAD that = (WIDE_DLOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_DLOAD'

    /** Defines the WIDE_DSTORE instruction: store double into local variable. */
    public static final class WIDE_DSTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_DSTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_DSTORE(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 57);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_DSTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 57;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_DSTORE) {
                final WIDE_DSTORE that = (WIDE_DSTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_DSTORE'

    /** Defines the WIDE_FLOAD instruction: load float from local variable. */
    public static final class WIDE_FLOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_FLOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_FLOAD(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 58);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_FLOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 58;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_FLOAD) {
                final WIDE_FLOAD that = (WIDE_FLOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_FLOAD'

    /** Defines the WIDE_FSTORE instruction: store float into local variable. */
    public static final class WIDE_FSTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_FSTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_FSTORE(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 59);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_FSTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 59;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_FSTORE) {
                final WIDE_FSTORE that = (WIDE_FSTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_FSTORE'

    /** Defines the WIDE_IINC instruction: increment local variable by constant. */
    public static final class WIDE_IINC extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /** Value of second instruction parameter. */
        private final short p2;

        /**
         * Constructs an instance of the WIDE_IINC instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         * @param p2 value of second instruction parameter
         */
        public WIDE_IINC(final int p1, final short p2) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
            this.p2 = p2;
        } // end constructor (int, short)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * Returns the value of the second parameter.
         * @return the value of the second parameter
         */
        public short getSecondParameter() {
            return this.p2;
        } // end method 'getSecondParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(2, 60);
            b.set(0, Helper.createInt(this.p1));
            b.set(1, Helper.createInt(this.p2));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_IINC" + " " + this.p1 + " " + this.p2;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 60;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_IINC) {
                final WIDE_IINC that = (WIDE_IINC) obj;
                return this.p1 == that.p1
                    && this.p2 == that.p2;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_IINC'

    /** Defines the WIDE_ILOAD instruction: load int from local variable. */
    public static final class WIDE_ILOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_ILOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_ILOAD(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 61);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_ILOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 61;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_ILOAD) {
                final WIDE_ILOAD that = (WIDE_ILOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_ILOAD'

    /** Defines the WIDE_ISTORE instruction: store int into local variable. */
    public static final class WIDE_ISTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_ISTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_ISTORE(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 62);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_ISTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 62;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_ISTORE) {
                final WIDE_ISTORE that = (WIDE_ISTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_ISTORE'

    /** Defines the WIDE_LLOAD instruction: load long from local variable. */
    public static final class WIDE_LLOAD extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_LLOAD instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_LLOAD(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 63);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_LLOAD" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 63;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_LLOAD) {
                final WIDE_LLOAD that = (WIDE_LLOAD) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_LLOAD'

    /** Defines the WIDE_LSTORE instruction: store long into local variable. */
    public static final class WIDE_LSTORE extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_LSTORE instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_LSTORE(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 64);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_LSTORE" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 64;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_LSTORE) {
                final WIDE_LSTORE that = (WIDE_LSTORE) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_LSTORE'

    /** Defines the WIDE_RET instruction: return from subroutine. */
    public static final class WIDE_RET extends Instruction {

        /** Value of first instruction parameter. */
        private final int p1;

        /**
         * Constructs an instance of the WIDE_RET instruction.
         * @param p1 value of first instruction parameter - should be in 0..65535
         */
        public WIDE_RET(final int p1) {
            assert p1 >= 0 && p1 <= 65535 : "invalid unsigned short";
            this.p1 = p1;
        } // end constructor (int)

        /**
         * Returns the value of the first parameter.
         * @return the value of the first parameter
         */
        public int getFirstParameter() {
            return this.p1;
        } // end method 'getFirstParameter()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, 65);
            b.set(0, Helper.createInt(this.p1));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "WIDE_RET" + " " + this.p1;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return 65;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof WIDE_RET) {
                final WIDE_RET that = (WIDE_RET) obj;
                return this.p1 == that.p1;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'WIDE_RET'

    /** The tag for 'type' elements. */
    private static final int TAG_TYPE = Hash.hashVariant("Class_or_interface").asLong();

    /** The tag for 'array' elements. */
    private static final int TAG_ARRAY = Hash.hashVariant("Array_type").asLong();

    /** The tag for 'int' elements. */
    private static final int TAG_INT = Hash.hashVariant("Int").asLong();

    /** The tag for 'float' elements. */
    private static final int TAG_FLOAT = Hash.hashVariant("Float").asLong();

    /** The tag for 'double' elements. */
    private static final int TAG_DOUBLE = Hash.hashVariant("Double").asLong();

    /** The tag for 'long' elements. */
    private static final int TAG_LONG = Hash.hashVariant("Long").asLong();

    /** The tag for 'string' elements. */
    private static final int TAG_STRING = Hash.hashVariant("String").asLong();

    /**
     * Constructs an instance from a value.
     * @param v value to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance
     */
    @SuppressWarnings("fallthrough")
    static Instruction fromValue(final Value v) {
        assert v != null : "null v";
        if (v.isLong()) {
            switch (v.asLong()) {
            case 0: return AALOAD;
            case 1: return AASTORE;
            case 2: return ACONST_NULL;
            case 3: return ALOAD_0;
            case 4: return ALOAD_1;
            case 5: return ALOAD_2;
            case 6: return ALOAD_3;
            case 7: return ARETURN;
            case 8: return ARRAYLENGTH;
            case 9: return ASTORE_0;
            case 10: return ASTORE_1;
            case 11: return ASTORE_2;
            case 12: return ASTORE_3;
            case 13: return ATHROW;
            case 14: return BALOAD;
            case 15: return BASTORE;
            case 16: return CALOAD;
            case 17: return CASTORE;
            case 18: return D2F;
            case 19: return D2I;
            case 20: return D2L;
            case 21: return DADD;
            case 22: return DALOAD;
            case 23: return DASTORE;
            case 24: return DCMPG;
            case 25: return DCMPL;
            case 26: return DCONST_0;
            case 27: return DCONST_1;
            case 28: return DDIV;
            case 29: return DLOAD_0;
            case 30: return DLOAD_1;
            case 31: return DLOAD_2;
            case 32: return DLOAD_3;
            case 33: return DMUL;
            case 34: return DNEG;
            case 35: return DREM;
            case 36: return DRETURN;
            case 37: return DSTORE_0;
            case 38: return DSTORE_1;
            case 39: return DSTORE_2;
            case 40: return DSTORE_3;
            case 41: return DSUB;
            case 42: return DUP;
            case 43: return DUP2;
            case 44: return DUP2_X1;
            case 45: return DUP2_X2;
            case 46: return DUP_X1;
            case 47: return DUP_X2;
            case 48: return F2D;
            case 49: return F2I;
            case 50: return F2L;
            case 51: return FADD;
            case 52: return FALOAD;
            case 53: return FASTORE;
            case 54: return FCMPG;
            case 55: return FCMPL;
            case 56: return FCONST_0;
            case 57: return FCONST_1;
            case 58: return FCONST_2;
            case 59: return FDIV;
            case 60: return FLOAD_0;
            case 61: return FLOAD_1;
            case 62: return FLOAD_2;
            case 63: return FLOAD_3;
            case 64: return FMUL;
            case 65: return FNEG;
            case 66: return FREM;
            case 67: return FRETURN;
            case 68: return FSTORE_0;
            case 69: return FSTORE_1;
            case 70: return FSTORE_2;
            case 71: return FSTORE_3;
            case 72: return FSUB;
            case 73: return I2B;
            case 74: return I2C;
            case 75: return I2D;
            case 76: return I2F;
            case 77: return I2L;
            case 78: return I2S;
            case 79: return IADD;
            case 80: return IALOAD;
            case 81: return IAND;
            case 82: return IASTORE;
            case 83: return ICONST_0;
            case 84: return ICONST_1;
            case 85: return ICONST_2;
            case 86: return ICONST_3;
            case 87: return ICONST_4;
            case 88: return ICONST_5;
            case 89: return ICONST_M1;
            case 90: return IDIV;
            case 91: return ILOAD_0;
            case 92: return ILOAD_1;
            case 93: return ILOAD_2;
            case 94: return ILOAD_3;
            case 95: return IMUL;
            case 96: return INEG;
            case 97: return IOR;
            case 98: return IREM;
            case 99: return IRETURN;
            case 100: return ISHL;
            case 101: return ISHR;
            case 102: return ISTORE_0;
            case 103: return ISTORE_1;
            case 104: return ISTORE_2;
            case 105: return ISTORE_3;
            case 106: return ISUB;
            case 107: return IUSHR;
            case 108: return IXOR;
            case 109: return L2D;
            case 110: return L2F;
            case 111: return L2I;
            case 112: return LADD;
            case 113: return LALOAD;
            case 114: return LAND;
            case 115: return LASTORE;
            case 116: return LCMP;
            case 117: return LCONST_0;
            case 118: return LCONST_1;
            case 119: return LDIV;
            case 120: return LLOAD_0;
            case 121: return LLOAD_1;
            case 122: return LLOAD_2;
            case 123: return LLOAD_3;
            case 124: return LMUL;
            case 125: return LNEG;
            case 126: return LOR;
            case 127: return LREM;
            case 128: return LRETURN;
            case 129: return LSHL;
            case 130: return LSHR;
            case 131: return LSTORE_0;
            case 132: return LSTORE_1;
            case 133: return LSTORE_2;
            case 134: return LSTORE_3;
            case 135: return LSUB;
            case 136: return LUSHR;
            case 137: return LXOR;
            case 138: return MONITORENTER;
            case 139: return MONITOREXIT;
            case 140: return NOP;
            case 141: return POP;
            case 142: return POP2;
            case 143: return RETURN;
            case 144: return SALOAD;
            case 145: return SASTORE;
            case 146: return SWAP;
            default:
                assert false : "invalid tag";
                return null;
            } // end switch
        } else {
            final Block b = v.asBlock();
            switch (b.getTag()) {
            case 0: return new ALOAD((short) b.get(0).asLong());
            case 1: if (b.get(0).asBlock().get(0).asLong() == TAG_TYPE) return new ANEWARRAY_TYPE(TypeName.fromValue(b.get(0).asBlock().get(1))); else if (b.get(0).asBlock().get(0).asLong() == TAG_ARRAY) return new ANEWARRAY_ARRAY((Descriptor.Array) Descriptor.fromValue(b.get(0).asBlock().get(1)));
            case 2: return new ASTORE((short) b.get(0).asLong());
            case 3: return new BIPUSH((byte) b.get(0).asLong());
            case 4: if (b.get(0).asBlock().get(0).asLong() == TAG_TYPE) return new CHECKCAST_TYPE(TypeName.fromValue(b.get(0).asBlock().get(1))); else if (b.get(0).asBlock().get(0).asLong() == TAG_ARRAY) return new CHECKCAST_ARRAY((Descriptor.Array) Descriptor.fromValue(b.get(0).asBlock().get(1)));
            case 5: return new DLOAD((short) b.get(0).asLong());
            case 6: return new DSTORE((short) b.get(0).asLong());
            case 7: return new FLOAD((short) b.get(0).asLong());
            case 8: return new FSTORE((short) b.get(0).asLong());
            case 9: return new GETFIELD(FieldRef.fromValue(b.get(0)));
            case 10: return new GETSTATIC(FieldRef.fromValue(b.get(0)));
            case 11: return new GOTO((short) b.get(0).asLong());
            case 12: return new GOTO_W(b.get(0).asBlock().asInt32());
            case 13: return new IF_ACMPEQ((short) b.get(0).asLong());
            case 14: return new IF_ACMPNE((short) b.get(0).asLong());
            case 15: return new IF_ICMPEQ((short) b.get(0).asLong());
            case 16: return new IF_ICMPGE((short) b.get(0).asLong());
            case 17: return new IF_ICMPGT((short) b.get(0).asLong());
            case 18: return new IF_ICMPLE((short) b.get(0).asLong());
            case 19: return new IF_ICMPLT((short) b.get(0).asLong());
            case 20: return new IF_ICMPNE((short) b.get(0).asLong());
            case 21: return new IFEQ((short) b.get(0).asLong());
            case 22: return new IFGE((short) b.get(0).asLong());
            case 23: return new IFGT((short) b.get(0).asLong());
            case 24: return new IFLE((short) b.get(0).asLong());
            case 25: return new IFLT((short) b.get(0).asLong());
            case 26: return new IFNE((short) b.get(0).asLong());
            case 27: return new IFNONNULL((short) b.get(0).asLong());
            case 28: return new IFNULL((short) b.get(0).asLong());
            case 29: return new IINC((short) b.get(0).asLong(), (byte) b.get(1).asLong());
            case 30: return new ILOAD((short) b.get(0).asLong());
            case 31: if (b.get(0).asBlock().get(0).asLong() == TAG_TYPE) return new INSTANCEOF_TYPE(TypeName.fromValue(b.get(0).asBlock().get(1))); else if (b.get(0).asBlock().get(0).asLong() == TAG_ARRAY) return new INSTANCEOF_ARRAY((Descriptor.Array) Descriptor.fromValue(b.get(0).asBlock().get(1)));
            case 32: return new INVOKEDYNAMIC(DynamicMethodRef.fromValue(b.get(0)));
            case 33: return new INVOKEINTERFACE(MethodRef.fromValue(b.get(0)), (short) b.get(1).asLong());
            case 34: return new INVOKESPECIAL(MethodRef.fromValue(b.get(0)));
            case 35: return new INVOKESTATIC(MethodRef.fromValue(b.get(0)));
            case 36: if (b.get(0).asBlock().get(0).asBlock().get(0).asLong() == TAG_TYPE) return new INVOKEVIRTUAL(MethodRef.fromTaggedValue(b.get(0))); else if (b.get(0).asBlock().get(0).asBlock().get(0).asLong() == TAG_ARRAY) return new INVOKEVIRTUAL_ARRAY(ArrayMethodRef.fromTaggedValue(b.get(0)));
            case 37: return new ISTORE((short) b.get(0).asLong());
            case 38: return new JSR((short) b.get(0).asLong());
            case 39: return new JSR_W(b.get(0).asBlock().asInt32());
            case 40: if (b.get(0).asBlock().get(0).asLong() == TAG_INT) return new LDC_INT(b.get(0).asBlock().get(1).asBlock().asInt32()); else if (b.get(0).asBlock().get(0).asLong() == TAG_FLOAT) return new LDC_FLOAT((float) b.get(0).asBlock().get(1).asBlock().asDouble()); else if (b.get(0).asBlock().get(0).asLong() == TAG_STRING) return new LDC_STRING(b.get(0).asBlock().get(1).asBlock().asString()); else if (b.get(0).asBlock().get(0).asLong() == TAG_TYPE) return new LDC_TYPE(TypeName.fromValue(b.get(0).asBlock().get(1))); else if (b.get(0).asBlock().get(0).asLong() == TAG_ARRAY) return new LDC_ARRAY((Descriptor.Array) Descriptor.fromValue(b.get(0).asBlock().get(1)));
            case 41: if (b.get(0).asBlock().get(0).asLong() == TAG_LONG) return new LDC2_W_LONG(b.get(0).asBlock().get(1).asBlock().asInt64()); else if (b.get(0).asBlock().get(0).asLong() == TAG_DOUBLE) return new LDC2_W_DOUBLE(b.get(0).asBlock().get(1).asBlock().asDouble());
            case 42: if (b.get(0).asBlock().get(0).asLong() == TAG_INT) return new LDC_W_INT(b.get(0).asBlock().get(1).asBlock().asInt32()); else if (b.get(0).asBlock().get(0).asLong() == TAG_FLOAT) return new LDC_W_FLOAT((float) b.get(0).asBlock().get(1).asBlock().asDouble()); else if (b.get(0).asBlock().get(0).asLong() == TAG_STRING) return new LDC_W_STRING(b.get(0).asBlock().get(1).asBlock().asString()); else if (b.get(0).asBlock().get(0).asLong() == TAG_TYPE) return new LDC_W_TYPE(TypeName.fromValue(b.get(0).asBlock().get(1))); else if (b.get(0).asBlock().get(0).asLong() == TAG_ARRAY) return new LDC_W_ARRAY((Descriptor.Array) Descriptor.fromValue(b.get(0).asBlock().get(1)));
            case 43: return new LLOAD((short) b.get(0).asLong());
            case 44: return new LOOKUPSWITCH(b.get(0).asBlock().asInt32(), b.get(1).asBlock().asInt32(), MatchOffsetPair.listFromValue(b.get(2)));
            case 45: return new LSTORE((short) b.get(0).asLong());
            case 46: if (b.get(0).asBlock().get(0).asLong() == TAG_TYPE) return new MULTIANEWARRAY_TYPE(TypeName.fromValue(b.get(0).asBlock().get(1)), (short) b.get(0).asBlock().get(2).asLong()); else if (b.get(0).asBlock().get(0).asLong() == TAG_ARRAY) return new MULTIANEWARRAY_ARRAY((Descriptor.Array) Descriptor.fromValue(b.get(0).asBlock().get(1)), (short) b.get(0).asBlock().get(2).asLong());
            case 47: return new NEW(TypeName.fromValue(b.get(0)));
            case 48: return new NEWARRAY(Descriptor.fromValue(b.get(0)));
            case 49: return new PUTFIELD(FieldRef.fromValue(b.get(0)));
            case 50: return new PUTSTATIC(FieldRef.fromValue(b.get(0)));
            case 51: return new RET((short) b.get(0).asLong());
            case 52: return new SIPUSH((short) b.get(0).asLong());
            case 53: return new TABLESWITCH(b.get(0).asBlock().asInt32(), b.get(1).asBlock().asInt32(), b.get(2).asBlock().asInt32(), readIntegerList(b.get(3)));
            case 54: return new WIDE_ALOAD(b.get(0).asLong());
            case 55: return new WIDE_ASTORE(b.get(0).asLong());
            case 56: return new WIDE_DLOAD(b.get(0).asLong());
            case 57: return new WIDE_DSTORE(b.get(0).asLong());
            case 58: return new WIDE_FLOAD(b.get(0).asLong());
            case 59: return new WIDE_FSTORE(b.get(0).asLong());
            case 60: return new WIDE_IINC(b.get(0).asLong(), (short) b.get(1).asLong());
            case 61: return new WIDE_ILOAD(b.get(0).asLong());
            case 62: return new WIDE_ISTORE(b.get(0).asLong());
            case 63: return new WIDE_LLOAD(b.get(0).asLong());
            case 64: return new WIDE_LSTORE(b.get(0).asLong());
            case 65: return new WIDE_RET(b.get(0).asLong());
            default:
                assert false : "invalid tag";
                return null;
            } // end switch
        } // end if/else
    } // end method 'fromValue(Value)'

} // end class 'Instruction'
