(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Types *)

type t =
  [ `Public
  | `Private
  | `Protected
  | `Static
  | `Final
  | `Super
  | `Synchronized
  | `Bridge
  | `Volatile
  | `Transient
  | `Varargs
  | `Native
  | `Interface
  | `Abstract
  | `Strict
  | `Synthetic
  | `Annotation
  | `Enum ]

type for_class =
  [ `Public
  | `Final
  | `Super
  | `Interface
  | `Abstract
  | `Synthetic
  | `Annotation
  | `Enum ]

type for_inner_class =
  [ `Public
  | `Private
  | `Protected
  | `Static
  | `Final
  | `Super
  | `Interface
  | `Abstract
  | `Synthetic
  | `Annotation
  | `Enum ]

type for_field =
  [ `Public
  | `Private
  | `Protected
  | `Static
  | `Final
  | `Volatile
  | `Transient
  | `Synthetic
  | `Enum ]

type for_method =
  [ `Public
  | `Private
  | `Protected
  | `Static
  | `Final
  | `Synchronized
  | `Bridge
  | `Varargs
  | `Native
  | `Abstract
  | `Strict
  | `Synthetic ]

type for_constructor =
  [ `Public
  | `Private
  | `Protected
  | `Strict
  | `Varargs
  | `Synthetic ]


(* Exception *)

type error =
  | Invalid_class_flags
  | Invalid_inner_class_flags
  | Invalid_field_flags
  | Invalid_method_flags
  | Invalid_constructor_flags

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_class_flags -> "invalid flags for class"
  | Invalid_inner_class_flags -> "invalid flags for inner class"
  | Invalid_field_flags -> "invalid flags for field"
  | Invalid_method_flags -> "invalid flags for method"
  | Invalid_constructor_flags -> "invalid flags for constructor"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* Functions *)

let all = [
  `Public ;
  `Private ;
  `Protected ;
  `Static ;
  `Final ;
  `Super ;
  `Volatile ;
  `Transient ;
  `Native ;
  `Interface ;
  `Abstract ;
  `Strict ;
  `Synthetic ;
  `Annotation ;
  `Enum
]

let all_for_method = [
  `Public ;
  `Private ;
  `Protected ;
  `Static ;
  `Final ;
  `Synchronized ;
  `Bridge ;
  `Varargs ;
  `Native ;
  `Interface ;
  `Abstract ;
  `Strict ;
  `Synthetic ;
  `Annotation ;
  `Enum
]

let to_u2 = function
  | `Public -> 0x0001
  | `Private -> 0x0002
  | `Protected -> 0x0004
  | `Static -> 0x0008
  | `Final -> 0x0010
  | `Super -> 0x0020
  | `Synchronized -> 0x0020
  | `Bridge -> 0x0040
  | `Volatile -> 0x0040
  | `Transient -> 0x0080
  | `Varargs -> 0x0080
  | `Native -> 0x0100
  | `Interface -> 0x0200
  | `Abstract -> 0x0400
  | `Strict -> 0x0800
  | `Synthetic -> 0x1000
  | `Annotation -> 0x2000
  | `Enum -> 0x4000

let list_to_u2 l =
  let res = ref 0 in
  let rec iter = function
    | hd :: tl -> res := !res + (to_u2 hd); iter tl
    | [] -> () in
  iter l;
  !res

let from_u2 meth mask =
  List.fold_left
    (fun l e -> if mask land (to_u2 e) <> 0 then e :: l else l)
    []
    (if meth then all_for_method else all)

let check_visibility fl =
  let incr_if x = if !x = 0 then incr x in
  let pub = ref 0 in
  let pro = ref 0 in
  let pri = ref 0 in
  let iter = function
    | `Public -> incr_if pub
    | `Protected -> incr_if pro
    | `Private -> incr_if pri
    | _ -> () in
  List.iter iter fl;
  (!pub + !pro + !pri) <= 1

let implies l =
  fun x y -> not (List.mem x l) || (List.mem y l)

let implies_not l =
  fun x y -> not (List.mem x l) || not (List.mem y l)

let check_class_flags fl =
  let ( ==> ) = implies fl in
  let ( =/> ) = implies_not fl in
  if ((List.fold_left
         (fun a f -> if f = `Public then succ a else a)
         0
         fl) <= 1)
      && (`Interface ==> `Abstract)
      && (`Interface =/> `Final)
      && (`Interface =/> `Super)
      && (`Interface =/> `Enum)
      && (`Annotation ==> `Interface)
      && (`Final =/> `Abstract)
      && (`Abstract =/> `Final)
  then List.map (function #for_class as x -> x | _ -> fail Invalid_class_flags) fl
  else fail Invalid_class_flags

let check_inner_class_flags fl =
  let ( ==> ) = implies fl in
  let ( =/> ) = implies_not fl in
  if (check_visibility fl)
      && (`Interface ==> `Abstract)
      && (`Interface =/> `Final)
      && (`Interface =/> `Super)
      && (`Interface =/> `Enum)
      && (`Annotation ==> `Interface)
      && (`Final =/> `Abstract)
      && (`Abstract =/> `Final)
  then List.map (function #for_inner_class as x -> x | _ -> fail Invalid_inner_class_flags) fl
  else fail Invalid_inner_class_flags

let check_field_flags interface fl =
  let ( =/> ) = implies_not fl in
  if (check_visibility fl)
      && (`Final =/> `Volatile)
      && (`Volatile =/> `Final)
      && ((not interface)
        || ((List.mem `Public fl)
              && (List.mem `Static fl)
              && (List.mem `Final fl)
              && (List.for_all (fun x -> List.mem x [`Public;
                                                     `Static;
                                                     `Final;
                                                     `Synthetic]) fl)))
  then List.map (function #for_field as x -> x | _ -> fail Invalid_field_flags) fl
  else fail Invalid_field_flags

let check_method_flags interface fl =
  let ( =/> ) = implies_not fl in
  if (check_visibility fl)
      && ((not interface)
        || ((List.mem `Public fl)
              && (List.mem `Abstract fl)
              && (List.for_all (fun x -> List.mem x [`Public;
                                                     `Abstract;
                                                     `Varargs;
                                                     `Bridge;
                                                     `Synthetic]) fl)))
      && (interface
        || ((`Abstract =/> `Final)
              && (`Abstract =/> `Native)
              && (`Abstract =/> `Private)
              && (`Abstract =/> `Static)
              && (`Abstract =/> `Strict)
              && (`Abstract =/> `Synchronized)))
  then List.map (function #for_method as x -> x | _ -> fail Invalid_method_flags) fl
  else fail Invalid_method_flags

let check_constructor_flags fl =
  if (check_visibility fl)
  then List.map (function #for_constructor as x -> x | _ -> fail Invalid_constructor_flags) fl
  else fail Invalid_constructor_flags

let version_bounds = function
  | `Public ->
      Version.make_bounds "'public' flag" Version.Java_1_0 None
  | `Private ->
      Version.make_bounds "'private' flag" Version.Java_1_0 None
  | `Protected ->
      Version.make_bounds "'protected' flag" Version.Java_1_0 None
  | `Static ->
      Version.make_bounds "'static' flag" Version.Java_1_0 None
  | `Final ->
      Version.make_bounds "'final' flag" Version.Java_1_0 None
  | `Super ->
      Version.make_bounds "'super' flag" Version.Java_1_0 None
  | `Synchronized ->
      Version.make_bounds "'synchronized' flag" Version.Java_1_0 None
  | `Bridge ->
      Version.make_bounds "'bridge' flag" Version.Java_1_5 None
  | `Volatile ->
      Version.make_bounds "'volatile' flag" Version.Java_1_0 None
  | `Transient ->
      Version.make_bounds "'transient' flag" Version.Java_1_0 None
  | `Varargs ->
      Version.make_bounds "'varargs' flag" Version.Java_1_5 None
  | `Native ->
      Version.make_bounds "'native' flag" Version.Java_1_0 None
  | `Interface ->
      Version.make_bounds "'interface' flag" Version.Java_1_0 None
  | `Abstract ->
      Version.make_bounds "'abstract' flag" Version.Java_1_0 None
  | `Strict ->
      Version.make_bounds "'strict' flag" Version.Java_1_1 None
  | `Synthetic ->
      Version.make_bounds "'synthetic' flag" Version.Java_1_5 None
  | `Annotation ->
      Version.make_bounds "'annotation' flag" Version.Java_1_5 None
  | `Enum ->
      Version.make_bounds "'enum' flag" Version.Java_1_5 None
