(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** This module defines access flags for various Java elements. *)


(** {6 Types} *)

type t =
  [ `Public
  | `Private
  | `Protected
  | `Static
  | `Final
  | `Super
  | `Synchronized
  | `Bridge
  | `Volatile
  | `Transient
  | `Varargs
  | `Native
  | `Interface
  | `Abstract
  | `Strict
  | `Synthetic
  | `Annotation
  | `Enum ]
(** Possible flags for any Java element. *)

type for_class =
  [ `Public
  | `Final
  | `Super
  | `Interface
  | `Abstract
  | `Synthetic
  | `Annotation
  | `Enum ]
(** Possible flags for a class. *)

type for_inner_class =
  [ `Public
  | `Private
  | `Protected
  | `Static
  | `Final
  | `Super
  | `Interface
  | `Abstract
  | `Synthetic
  | `Annotation
  | `Enum ]
(** Possible flags for an inner-class. *)

type for_field =
  [ `Public
  | `Private
  | `Protected
  | `Static
  | `Final
  | `Volatile
  | `Transient
  | `Synthetic
  | `Enum ]
(** Possible flags for a field. *)

type for_method =
  [ `Public
  | `Private
  | `Protected
  | `Static
  | `Final
  | `Synchronized
  | `Bridge
  | `Varargs
  | `Native
  | `Abstract
  | `Strict
  | `Synthetic ]
(** Possible flags for a method. *)

type for_constructor =
  [ `Public
  | `Private
  | `Protected
  | `Strict
  | `Varargs
  | `Synthetic ]
(** Possible flags for a constructor. *)


(** {6 Exception} *)

type error =
  | Invalid_class_flags
  | Invalid_inner_class_flags
  | Invalid_field_flags
  | Invalid_method_flags
  | Invalid_constructor_flags

exception Exception of error
(** Exception to be raised if a flag is incorrect. *)

val string_of_error : error -> string
(** Converts the passed error into a string. *)


(** {6 Functions} *)

val to_u2 : t -> Utils.u2
(** [to_u2 f] converts flag [f] into an integer mask,
    as used in the class file format. *)

val list_to_u2 : t list -> Utils.u2
(** [to_u2 dom l] converts flag list [l] into an integer mask,
    as used in the class file format. *)

val from_u2 : bool -> Utils.u2 -> t list
(** [from_u2 meth mask] converts mask into flag list, [meth] indicating whether
   the considered Java element is a method (including constructors). *)

val check_class_flags : t list -> for_class list
(** Acts as the identity function if the passed flags form a valid flag set
   for a class, raises [Exception] otherwise. *)

val check_inner_class_flags : t list -> for_inner_class list
(** Acts as the identity function if the passed flags form a valid flag set
   for an inner class, raises [Exception] otherwise. *)

val check_field_flags : bool -> t list -> for_field list
(** Acts as the identity function if the passed flags form a valid flag set
   for a field, raises [Exception] otherwise.
   The passed boolean indicates whether the checked field belongs to an
   interface. *)

val check_method_flags : bool -> t list -> for_method list
(** Acts as the identity function if the passed flags form a valid flag set
   for a method, raises [Exception] otherwise.
   The passed boolean indicates whether the checked method belongs to an
   interface. *)

val check_constructor_flags : t list -> for_constructor list
(** Acts as the identity function if the passed flags form a valid flag set
   for a constructor, raises [Exception] otherwise. *)

val version_bounds : t -> Version.bounds
(** Returns the version bounds for the passed flag. *)
