(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils


type t = {
    class_path : ClassPath.t;
    loaded_classes : ClassDefinition.t UTF8Hashtbl.t;
  }

type error =
  | Unable_to_load of utf8 * (string option)
  | Already_defined

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Unable_to_load (s, os) ->
      Printf.sprintf
      "unable to load class %s%s"
        (string_of_utf8 s)
        (match os with Some v -> " (" ^ v ^ ")" | None -> "")
  | Already_defined -> "class is already defined"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)

let make cp =
  { class_path = cp;
    loaded_classes = UTF8Hashtbl.create 16; }

let find cl cn =
  try
    UTF8Hashtbl.find cl.loaded_classes cn
  with Not_found ->
    let stream = ClassPath.open_stream cl.class_path (string_of_utf8 cn) in
    try
      let cf = ClassFile.read stream in
      let cd = ClassDefinition.decode cf in
      UTF8Hashtbl.add cl.loaded_classes cn cd;
      cd
    with (ClassFile.Exception cause) ->
      (try InputStream.close stream with _ -> ());
      fail (Unable_to_load (cn, Some (ClassFile.string_of_error cause)))
    | _ ->
      (try InputStream.close stream with _ -> ());
      fail (Unable_to_load (cn, None))

let add cl cd =
  let cn = Name.external_utf8_for_class cd.ClassDefinition.name in
  if UTF8Hashtbl.mem cl.loaded_classes cn then
    fail Already_defined
  else
    UTF8Hashtbl.add cl.loaded_classes cn cd
