(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils


(* Low-level form *)

type info = {
    access_flags : u2;
    name_index : u2;
    descriptor_index : u2;
    attributes_count : u2;
    attributes : Attribute.info array;
  }


(* Exception *)

type error =
  | Invalid_name
  | Invalid_name_value
  | Invalid_descriptor_value

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_name -> "invalid name"
  | Invalid_name_value -> "invalid name value"
  | Invalid_descriptor_value -> "invalid descriptor value"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* I/O functions *)

let read_info st =
  let flags = InputStream.read_u2 st in
  let name = InputStream.read_u2 st in
  let desc = InputStream.read_u2 st in
  let att_count = InputStream.read_u2 st in
  let atts = Array.init att_count (fun _ -> Attribute.read_info st) in
  { access_flags = flags;
    name_index = name;
    descriptor_index = desc;
    attributes_count = att_count;
    attributes = atts; }

let write_info st i =
  OutputStream.write_u2 st i.access_flags;
  OutputStream.write_u2 st i.name_index;
  OutputStream.write_u2 st i.descriptor_index;
  OutputStream.write_u2 st i.attributes_count;
  Array.iter (Attribute.write_info st) i.attributes


(* High-level form *)

type t = (AccessFlag.for_field list) * Name.for_field * Descriptor.for_field * (Attribute.for_field list)


(* Conversion functions *)

let decode itf pool i =
  let flags = AccessFlag.check_field_flags itf (AccessFlag.from_u2 false i.access_flags) in
  let name = match ConstantPool.get_entry pool i.name_index with
  | ConstantPool.UTF8 n ->
      if Name.is_valid_unqualified n then
        n
      else
        fail Invalid_name
  | _ -> fail Invalid_name_value in
  let desc = match ConstantPool.get_entry pool i.descriptor_index with
  | ConstantPool.UTF8 d -> Descriptor.field_of_utf8 d
  | _ -> fail Invalid_descriptor_value in
  let attrs = List.map (Attribute.decode Attribute.Field pool) (Array.to_list i.attributes) in
  (flags, (Name.make_for_field name), desc, (Attribute.check_field_attributes attrs))

let encode pool f =
  let flags, name, desc, attrs = f in
  let acc_flags = AccessFlag.list_to_u2 (flags :> AccessFlag.t list) in
  let name_idx = ConstantPool.add_utf8 pool (Name.utf8_for_field name) in
  let desc_utf8 = Descriptor.utf8_of_field desc in
  let desc_idx = ConstantPool.add_utf8 pool desc_utf8 in
  { access_flags = acc_flags;
    name_index = name_idx;
    descriptor_index = desc_idx;
    attributes_count = List.length attrs;
    attributes = Array.of_list (List.map (Attribute.encode pool) (attrs :> Attribute.t list)); }
