(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Tokens *)

type token =
  | Directive of string
  | Attribute of string
  | Label of utf8
  | Int of int64
  | Float of float
  | String of utf8
  | Class_name of Name.for_class
  | Array_type of utf8
  | Primitive_type of Descriptor.java_type
  | Field of Name.for_class * Name.for_field * Descriptor.for_field
  | Dynamic_method of Name.for_method * Descriptor.for_method
  | Method of Name.for_class * Name.for_method * Descriptor.for_method
  | Array_method of Descriptor.array_type * Name.for_method * Descriptor.for_method
  | Method_signature of Name.for_method * (Descriptor.for_parameter list)
  | Identifier of utf8
  | Arrow
  | Tilde


(* Exception *)

type error =
  | Invalid_label
  | Invalid_directive
  | Invalid_attribute
  | Invalid_string
  | Invalid_character
  | Invalid_float
  | Invalid_integer
  | Invalid_token
  | Name_error of Name.error
  | Descriptor_error of Descriptor.error
  | Source_error of Source.error
  | Unicode_error of unicode_error

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_label -> "invalid label"
  | Invalid_directive -> "invalid directive"
  | Invalid_attribute -> "invalid attribute"
  | Invalid_string -> "invalid string"
  | Invalid_character -> "invalid character"
  | Invalid_float -> "invalid float constant"
  | Invalid_integer -> "invalid integer constant"
  | Invalid_token -> "invalid token"
  | Name_error e -> Name.string_of_error e
  | Descriptor_error e -> Descriptor.string_of_error e
  | Source_error e -> Source.string_of_error e
  | Unicode_error e -> string_of_unicode_error e

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* Lexing funtion *)

let analyse_token s =
  let len = utf8_length s in
  let last = pred len in
  if uchar_eq colon (utf8_get s last) then begin
    let i = ref 1 in
    while !i < last && uchar_is_letter_or_digit (utf8_get s !i) do
      incr i
    done;
    if !i = last && len > 1 && uchar_is_letter (utf8_get s 0) then
      Label s
    else
      fail Invalid_label
  end else if utf8_contains opening_parenthesis s && utf8_contains closing_parenthesis s then begin
    let opening_index = utf8_index_from s 0 opening_parenthesis in
    let closing_index = utf8_rindex_from s last closing_parenthesis in
    let prefix = utf8_substring s 0 (pred opening_index) in
    let params = utf8_substring s (succ opening_index) (pred closing_index) in
    if (succ closing_index) < len
        && uchar_eq colon (utf8_get s (succ closing_index)) then
      let desc = (List.map Source.java_type_of_utf8_no_void (utf8_split comma params)),
        (Source.java_type_of_utf8 (utf8_substring s (succ (succ closing_index)) last)) in
      if utf8_contains dot prefix then
        let dot_idx = utf8_rindex_from prefix (pred (utf8_length prefix)) dot in
        let class_name = utf8_substring prefix 0 (pred dot_idx) in
        let meth_name = utf8_substring prefix (succ dot_idx) (pred (utf8_length prefix)) in
        if utf8_contains opening_square_bracket class_name then
          Array_method ((Descriptor.filter_array Descriptor.Invalid_array_element_type (Source.java_type_of_utf8 class_name)),
                        (Name.make_for_method meth_name),
                        desc)
        else
          Method ((Name.make_for_class_from_external class_name),
                  (Name.make_for_method meth_name),
                  desc)
      else
        Dynamic_method ((Name.make_for_method prefix), desc)
    else
      Method_signature ((Name.make_for_method prefix),
                        (List.map Source.java_type_of_utf8_no_void (utf8_split comma params)))
  end else begin
    if utf8_contains colon s then
      let colon_idx = utf8_index_from s 0 colon in
      let prefix = utf8_substring s 0 (pred colon_idx) in
      let dot_idx = utf8_rindex_from prefix (pred (utf8_length prefix)) dot in
      Field ((Name.make_for_class_from_external (utf8_substring prefix 0 (pred dot_idx))),
             (Name.make_for_field (utf8_substring prefix (succ dot_idx) (pred (utf8_length prefix)))),
             (Source.java_type_of_utf8_no_void (utf8_substring s (succ colon_idx) (pred (utf8_length s)))))
    else if uchar_eq closing_square_bracket (utf8_get s last) then
      Array_type s
    else if utf8_contains dot s then
      Class_name (Name.make_for_class_from_external s)
    else
      try
        let t = Source.java_type_of_utf8 s in
        if (Descriptor.is_primitive t) || (t = `Void) then
            Primitive_type t
        else
          Identifier s
      with _ ->
        Identifier s
  end

let tokens_of_line l =
  let state = new lexer_state l in
  let skip_whitespace () =
    while state#is_available && state#look_ahead_list [space; tabulation] do
      state#consume
    done in
  let read_token () =
    let buf = utf8_buf_make () in
    if state#look_ahead dot then begin
      state#consume;
      while state#is_available && not (state#look_ahead_list [space; tabulation; sharp]) do
        utf8_buf_add_char buf state#consume_char
      done;
      let dir = string_of_utf8 (utf8_buf_contents buf) in
      if (String.length dir) = 0 then
        fail Invalid_directive
      else
        Directive dir
    end else if state#look_ahead at_character then begin
      state#consume;
      while state#is_available && not (state#look_ahead_list [space; tabulation; sharp]) do
        utf8_buf_add_char buf state#consume_char
      done;
      let attr = string_of_utf8 (utf8_buf_contents buf) in
      if (String.length attr) = 0 then
        fail Invalid_attribute
      else
        Attribute attr
    end else if state#look_ahead quote then begin
      let prev = ref state#consume_char in
      utf8_buf_add_char buf !prev;
      while not (state#look_ahead quote && not (uchar_eq !prev back_slash)) do
        let curr = state#consume_char in
        utf8_buf_add_char buf curr;
        prev :=
          if (uchar_eq !prev back_slash) && (uchar_eq curr back_slash) then
            quote
          else
            curr
      done;
      utf8_buf_add_char buf state#consume_char;
      if state#is_available && not (state#look_ahead_list [space; tabulation; sharp]) then
        fail Invalid_string
      else
         String (utf8_unescape (utf8_buf_contents buf))
    end else if state#look_ahead (uchar_of_char '\'') then begin
      let prev = ref state#consume_char in
      while not (state#look_ahead (uchar_of_char '\'') && not (uchar_eq !prev back_slash)) do
        let curr = state#consume_char in
        utf8_buf_add_char buf curr;
        prev :=
          if (uchar_eq !prev back_slash) && (uchar_eq curr back_slash) then
            quote
          else
            curr
      done;
      state#consume;
      if state#is_available && not (state#look_ahead_list [space; tabulation; sharp]) then fail Invalid_character;
      let s = utf8_unescape ((utf8_of_string "\"") ++ (utf8_buf_contents buf) ++ (utf8_of_string "\"")) in
      if (utf8_length s) <> 1 then
        fail Invalid_character
      else
        Int (Int64.of_int (code_of_uchar (utf8_get s 0)))
    end else if state#look_ahead_list ([minus; plus] @ digits) then begin
      while state#is_available && not (state#look_ahead_list [space; tabulation; sharp]) do
        utf8_buf_add_char buf state#consume_char
      done;
      let n = string_of_utf8 (utf8_buf_contents buf) in
      let number = if (String.get n 0) = '+' then String.sub n 1 (pred (String.length n)) else n in
      if String.contains number '.' then
        Float (float_of_string number)
      else
        Int (Int64.of_string number)
    end else if state#look_ahead (uchar_of_char '=') then begin
      state#consume;
      state#consume_only greater_than;
      Arrow
    end else if state#look_ahead (uchar_of_char '~') then begin
      state#consume;
      Tilde
    end else begin
      let continue = ref false in
      while state#is_available && not (state#look_ahead_list [space; tabulation; colon; sharp]) do
        let ch = state#consume_char in
        if uchar_eq dot ch then continue := true;
        utf8_buf_add_char buf ch;
        if uchar_eq opening_parenthesis ch then begin
          continue := true;
          while not (state#look_ahead closing_parenthesis) do
            let ch = state#consume_char in
            if not ((uchar_eq ch space) || (uchar_eq ch tabulation)) then
              utf8_buf_add_char buf ch
          done;
          utf8_buf_add_char buf state#consume_char;
        end
      done;
      skip_whitespace ();
      if state#is_available && (state#look_ahead colon) then begin
        utf8_buf_add_char buf state#consume_char;
        if !continue then begin
          skip_whitespace ();
          while state#is_available && not (state#look_ahead_list [space; tabulation; sharp]) do
            utf8_buf_add_char buf state#consume_char;
          done;
        end;
      end;
      analyse_token (utf8_buf_contents buf)
    end in
  let tokens = ref [] in
  while state#is_available do
    skip_whitespace ();
    if not state#is_available || state#look_ahead sharp then
      while state#is_available do state#consume done
    else
      let tok = try
        read_token ()
      with
      | Name.Exception e -> fail (Name_error e)
      | Descriptor.Exception e -> fail (Descriptor_error e)
      | Source.Exception e -> fail (Source_error e)
      | Unicode_exception e -> fail (Unicode_error e)
      | Exception e -> fail e
      | Failure "float_of_string" -> fail Invalid_float
      | Failure "int_of_string" -> fail Invalid_integer
      | _ -> fail Invalid_token in
      tokens := tok :: !tokens;
  done;
  List.rev !tokens


(* Miscellaneous *)

let eq x y =
  let rec type_desc_eq x y =
    match x, y with
    | (`Array a), (`Array a') ->
        type_desc_eq
        (a :> Descriptor.java_type)
        (a' :> Descriptor.java_type)
    | `Boolean, `Boolean
    | `Byte, `Byte
    | `Char, `Char
    | `Double, `Double
    | `Float, `Float
    | `Int, `Int
    | `Long, `Long
    | `Short, `Short
    | `Void, `Void -> true
    | (`Class c), (`Class c') ->
        utf8_eq
          (Name.internal_utf8_for_class c)
          (Name.internal_utf8_for_class c')
    | _ -> false in
  let method_desc_eq (p, r) (p', r') =
    (List.for_all2
       type_desc_eq
       (p :> Descriptor.java_type list)
       (p' :> Descriptor.java_type list))
      && (type_desc_eq r r') in
  let cls_eq x y =
    utf8_eq
      (Name.external_utf8_for_class x)
      (Name.external_utf8_for_class y) in
  let fld_eq x y =
    utf8_eq
      (Name.utf8_for_field x)
      (Name.utf8_for_field y) in
  let mth_eq x y =
    utf8_eq
      (Name.utf8_for_method x)
      (Name.utf8_for_method y) in
  match x, y with
  | (Directive d), (Directive d') -> d = d'
  | (Attribute a), (Attribute a') -> a = a'
  | (Label l), (Label l') -> utf8_eq l l'
  | (Int i), (Int i') -> i = i'
  | (Float f), (Float f') -> f = f'
  | (String s), (String s') -> utf8_eq s s'
  | (Class_name cn), (Class_name cn') -> cls_eq cn cn'
  | (Array_type at), (Array_type at') -> utf8_eq at at'
  | (Primitive_type pt), (Primitive_type pt') -> pt = pt'
  | (Field (x, y, z)), (Field (x', y', z')) ->
      (cls_eq x x')
        && (fld_eq y y')
        && (type_desc_eq
              (z :> Descriptor.java_type)
              (z' :> Descriptor.java_type))
  | (Dynamic_method (x, y)), (Dynamic_method (x', y')) ->
      (mth_eq x x') && (method_desc_eq y y')
  | (Method (x, y, z)), (Method (x', y', z')) ->
      (cls_eq x x') && (mth_eq y y') && (method_desc_eq z z')
  | (Array_method (x, y, z)), (Array_method (x', y', z')) ->
      (type_desc_eq
         (x :> Descriptor.java_type)
         (x' :> Descriptor.java_type))
        && (mth_eq y y')
        && (method_desc_eq z z')
  | (Method_signature (x, y)), (Method_signature (x', y')) ->
      (mth_eq x x')
        && (List.for_all2
              type_desc_eq
              (y :> Descriptor.java_type list)
              (y' :> Descriptor.java_type list))
  | (Identifier i), (Identifier i') -> utf8_eq i i'
  | Arrow, Arrow -> true
  | _ -> false
