(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** This module provides some utility functions related to name handling as
   well as definition of class, field and method names. *)


(** {6 Utility functions} *)

val replace_dot_with_slash : Utils.utf8 -> Utils.utf8
(** [replace_dot_with_slash s] returns a copy of string [s] where each dot has
   been replaced by a slash. Useful to transform a fully qualified class name
   from external format into internal format. *)

val replace_slash_with_dot : Utils.utf8 -> Utils.utf8
(** [replace_slash_with_dot s] returns a copy of string [s] where each slash
   and dollar has been replaced by a dot. Useful to transform a fully qualified
   class name from internal format into external format. *)

val is_valid_unqualified : Utils.utf8 -> bool
(** Checks whether the passed string is a valid unqualified name.
   That is, returns true iff the passed string is non-empty, does not contain any
   dot, semi colon, opening square bracket or slash. *)

val is_valid_for_method : Utils.utf8 -> bool
(** Checks whether the passed string is a valid method name.
   That is, returns true iff the passed string is non-empty, is either the class
   constructor/initializer or is a valid unqualified name that does not
   contain any '<' or '>'.*)


(** {6 Exception} *)

type error =
  | Invalid_class_name
  | Invalid_field_name
  | Invalid_method_name

exception Exception of error
(** Raised when a function of this module fails. *)

val string_of_error : error -> string
(** Converts the passed error into a string. *)


(** {6 Name types and conversion functions} *)

type for_class
(** The type of class names. *)

type for_field
(** The type of field names. *)

type for_method
(** The type of method names. *)

val make_for_class_from_internal : Utils.utf8 -> for_class
(** Constructs a class name from an utf8 string (slash being the separator
   between package element, dollar being the separator between inner elements).
   Raises [Exception] if passed utf8 is invalid. *)

val make_for_class_from_external : Utils.utf8 -> for_class
(** Constructs a class name from an utf8 string (dot being the separator
   between package element, dollar being the separator between inner elements).
   Raises [Exception] if passed utf8 is invalid. *)

val make_for_field : Utils.utf8 -> for_field
(** Constructs a field name from an utf8 string.
   Raises [Exception] if passed utf8 is invalid. *)

val make_for_method : Utils.utf8 -> for_method
(** Constructs a method name from an utf8 string.
   Raises [Exception] if passed utf8 is invalid. *)

val printable_utf8_for_class : for_class -> Utils.utf8
(** Converts a class name into external utf8 form
   (dots between both package and inner elements). *)

val external_utf8_for_class : for_class -> Utils.utf8
(** Converts a class name into external utf8 form
   (dots between package elements, dollars between inner elements). *)

val internal_utf8_for_class : for_class -> Utils.utf8
(** Converts a class name into internal utf8 form.
   (slashes between package elements, dollars between inner elements). *)

val utf8_for_field : for_field -> Utils.utf8
(** Converts a field name into utf8 form. *)

val utf8_for_method : for_method -> Utils.utf8
(** Converts a method name into utf8 form. *)
