/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.FileOutputStream;
import java.util.Collections;
import java.util.List;
import static java.util.Arrays.asList;

import fr.x9c.barista.api.*;
import static fr.x9c.barista.api.Instruction.*;
import static tests.Common.*;

public final class Source {

    private static final FieldRef systOut =
	new FieldRef("java.lang.System",
		     "out",
		     new Descriptor.Class("java.io.PrintStream"));

    private static final MethodRef println =
	new MethodRef("java.io.PrintStream",
		      "println",
		      Descriptor.VOID,
		      descs(new Descriptor.Class("java.lang.String")));

    private static final MethodRef toString =
	new MethodRef("java.lang.Object",
		      "toString",
		      new Descriptor.Class("java.lang.String"),
		      descs());

    public static void main(final String[] args) throws Throwable {
	final Field n = new Field(asList(AccessFlag.Public,
					 AccessFlag.Static,
					 AccessFlag.Final),
				  "N",
				  Descriptor.INT,
				  Collections.<Attribute>singletonList(new Attribute.ConstantValue(new Attribute.IntegerValue(5))));
	final List<Instruction> floatsInstructions =
	    asList(new GETSTATIC(new FieldRef("pack.Test", "N", Descriptor.INT)),
		   new NEWARRAY(Descriptor.FLOAT),
		   DUP,
		   DUP,
		   ICONST_1,
		   FALOAD,
		   FCONST_2,
		   FADD,
		   ICONST_1,
		   SWAP,
		   FASTORE,
		   new INVOKESTATIC(new MethodRef("java.util.Arrays",
						  "toString",
						  new Descriptor.Class("java.lang.String"),
						  descs(new Descriptor.Array(Descriptor.FLOAT)))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> doublesInstructions =
	    asList(new GETSTATIC(new FieldRef("pack.Test", "N", Descriptor.INT)),
		   new NEWARRAY(Descriptor.DOUBLE),
		   DUP,
		   DUP,
		   ICONST_1,
		   DALOAD,
		   DCONST_1,
		   DADD,
		   DSTORE_0,
		   ICONST_1,
		   DLOAD_0,
		   DASTORE,
		   new INVOKESTATIC(new MethodRef("java.util.Arrays",
						  "toString",
						  new Descriptor.Class("java.lang.String"),
						  descs(new Descriptor.Array(Descriptor.DOUBLE)))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> intsInstructions =
	    asList(new GETSTATIC(new FieldRef("pack.Test", "N", Descriptor.INT)),
		   new NEWARRAY(Descriptor.INT),
		   DUP,
		   DUP,
		   ICONST_1,
		   IALOAD,
		   ICONST_3,
		   IADD,
		   ICONST_1,
		   SWAP,
		   IASTORE,
		   new INVOKESTATIC(new MethodRef("java.util.Arrays",
						  "toString",
						  new Descriptor.Class("java.lang.String"),
						  descs(new Descriptor.Array(Descriptor.INT)))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> longsInstructions =
	    asList(new GETSTATIC(new FieldRef("pack.Test", "N", Descriptor.INT)),
		   new NEWARRAY(Descriptor.LONG),
		   DUP,
		   DUP,
		   ICONST_2,
		   LALOAD,
		   LCONST_1,
		   LADD,
		   LSTORE_0,
		   ICONST_2,
		   LLOAD_0,
		   LASTORE,
		   new INVOKESTATIC(new MethodRef("java.util.Arrays",
						  "toString",
						  new Descriptor.Class("java.lang.String"),
						  descs(new Descriptor.Array(Descriptor.LONG)))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> bytesInstructions =
	    asList(new GETSTATIC(new FieldRef("pack.Test", "N", Descriptor.INT)),
		   new NEWARRAY(Descriptor.BYTE),
		   DUP,
		   DUP,
		   ICONST_1,
		   BALOAD,
		   ICONST_3,
		   IADD,
		   ICONST_1,
		   SWAP,
		   BASTORE,
		   new INVOKESTATIC(new MethodRef("java.util.Arrays",
						  "toString",
						  new Descriptor.Class("java.lang.String"),
						  descs(new Descriptor.Array(Descriptor.BYTE)))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> charsInstructions =
	    asList(new GETSTATIC(new FieldRef("pack.Test", "N", Descriptor.INT)),
		   new NEWARRAY(Descriptor.CHAR),
		   DUP,
		   DUP,
		   DUP,
		   ICONST_1,
		   CALOAD,
		   ICONST_3,
		   IADD,
		   ICONST_1,
		   SWAP,
		   CASTORE,
		   ICONST_2,
		   new LDC_INT(90),
		   CASTORE,
		   new INVOKESTATIC(new MethodRef("java.util.Arrays",
						  "toString",
						  new Descriptor.Class("java.lang.String"),
						  descs(new Descriptor.Array(Descriptor.CHAR)))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> shortsInstructions =
	    asList(new GETSTATIC(new FieldRef("pack.Test", "N", Descriptor.INT)),
		   new NEWARRAY(Descriptor.SHORT),
		   DUP,
		   DUP,
		   ICONST_1,
		   SALOAD,
		   ICONST_3,
		   IADD,
		   ICONST_1,
		   SWAP,
		   SASTORE,
		   new INVOKESTATIC(new MethodRef("java.util.Arrays",
						  "toString",
						  new Descriptor.Class("java.lang.String"),
						  descs(new Descriptor.Array(Descriptor.SHORT)))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> stringsInstructions =
	    asList(new GETSTATIC(new FieldRef("pack.Test", "N", Descriptor.INT)),
		   new ANEWARRAY_TYPE(new TypeName("java.lang.String")),
		   DUP,
		   DUP,
		   DUP,
		   DUP,
		   DUP,
		   ICONST_0,
		   new LDC_STRING ("first"),
		   AASTORE,
		   ICONST_1,
		   new LDC_STRING ("second"),
		   AASTORE,
		   ICONST_2,
		   new LDC_STRING ("third"),
		   AASTORE,
		   ICONST_3,
		   new LDC_STRING ("fourth"),
		   AASTORE,
		   ICONST_4,
		   new LDC_STRING ("fifth"),
		   AASTORE,
		   new INVOKESTATIC(new MethodRef("java.util.Arrays",
						  "toString",
						  new Descriptor.Class("java.lang.String"),
						  descs(new Descriptor.Array(new Descriptor.Class("java.lang.Object"))))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> matrixInstructions =
	    asList(new GETSTATIC(systOut),
		   new LDC_STRING("matrix:"),
		   new INVOKEVIRTUAL(println),
		   new GETSTATIC(new FieldRef("pack.Test", "N", Descriptor.INT)),
		   DUP,
		   new MULTIANEWARRAY_ARRAY(new Descriptor.Array(new Descriptor.Array(new Descriptor.Class("java.lang.String"))), (short) 2),
		   ASTORE_0,
		   ICONST_0,
		   ISTORE_1,
		   ALOAD_0,
		   ARRAYLENGTH,
		   ISTORE_2,
		   /* loop: */
		   ILOAD_1,
		   ILOAD_2,
		   new IF_ICMPEQ((short) 22),
		   ALOAD_0,
		   ILOAD_1,
		   AALOAD,
		   new INVOKESTATIC(new MethodRef("java.util.Arrays",
						  "toString",
						  new Descriptor.Class("java.lang.String"),
						  descs(new Descriptor.Array(new Descriptor.Class("java.lang.Object"))))),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   new IINC((short) 1, (byte) 1),
		   new GOTO((short) -21),
		   /* end: */
		   RETURN);
	final List<Instruction> classesInstructions =
	    asList(new GETSTATIC(systOut),
		   new LDC_ARRAY(new Descriptor.Array(Descriptor.INT)),
		   new INVOKEVIRTUAL(toString),
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> instructions =
	    asList(new INVOKESTATIC(new MethodRef("pack.Test",
						  "floats",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "doubles",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "ints",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "longs",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "bytes",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "chars",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "shorts",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "strings",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "matrix",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "classes",
						  Descriptor.VOID,
						  descs())),
		   RETURN);
	final Method floats =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "floats",
			  Descriptor.VOID,
			  descs(),
			  8,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  floatsInstructions);
	final Method doubles =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "doubles",
			  Descriptor.VOID,
			  descs(),
			  8,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  doublesInstructions);
	final Method ints =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "ints",
			  Descriptor.VOID,
			  descs(),
			  8,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  intsInstructions);
	final Method longs =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "longs",
			  Descriptor.VOID,
			  descs(),
			  8,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  longsInstructions);
	final Method bytes =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "bytes",
			  Descriptor.VOID,
			  descs(),
			  8,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  bytesInstructions);
	final Method chars =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "chars",
			  Descriptor.VOID,
			  descs(),
			  8,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  charsInstructions);
	final Method shorts =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "shorts",
			  Descriptor.VOID,
			  descs(),
			  8,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  shortsInstructions);
	final Method strings =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "strings",
			  Descriptor.VOID,
			  descs(),
			  8,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  stringsInstructions);
	final Method matrix =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "matrix",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  matrixInstructions);
	final Method classes =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "classes",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  classesInstructions);
	final Method main = compileMethod(instructions);
	final ClassDefinition cd = compileClass(null,
						null,
						null,
						null,
						asList(n),
						null,
						asList(floats,
						       doubles,
						       ints,
						       longs,
						       bytes,
						       chars,
						       shorts,
						       strings,
						       matrix,
                                                       classes,
						       main));
	final byte[] bc = ByteCode.encode(cd);
	FileOutputStream out = new FileOutputStream("pack/Test.class");
	out.write(bc);
	out.flush();
    }

}
